#!/usr/bin/env python
# supplier_finder.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (supplier_finder.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: supplier_finder.

.. role:: red
BitDust supplier_finder() Automat


EVENTS:
    * :red:`ack-received`
    * :red:`found-one-user`
    * :red:`ping-failed`
    * :red:`start`
    * :red:`supplier-connected`
    * :red:`supplier-not-connected`
    * :red:`timer-10sec`
    * :red:`timer-30sec`
    * :red:`users-not-found`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import strng

from bitdust.p2p import lookup
from bitdust.p2p import online_status

from bitdust.contacts import identitycache
from bitdust.contacts import contactsdb

from bitdust.userid import my_id
from bitdust.userid import id_url

#------------------------------------------------------------------------------

_SupplierFinder = None
_SuppliersToHire = []

#------------------------------------------------------------------------------


def AddSupplierToHire(idurl):
    global _SuppliersToHire
    idurl = strng.to_bin(idurl)
    if idurl not in _SuppliersToHire:
        _SuppliersToHire.append(idurl)
        if _Debug:
            lg.dbg(_DebugLevel, 'added %r as a supplier candidate' % idurl)


def InsertSupplierToHire(idurl):
    global _SuppliersToHire
    idurl = strng.to_bin(idurl)
    if idurl not in _SuppliersToHire:
        _SuppliersToHire.insert(0, idurl)
        if _Debug:
            lg.dbg(_DebugLevel, 'added %r as a FIRST supplier candidate' % idurl)


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _SupplierFinder
    if _SupplierFinder is None:
        # set automat name and starting state here
        _SupplierFinder = SupplierFinder(
            name='supplier_finder',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _SupplierFinder.automat(event, *args, **kwargs)
    return _SupplierFinder


class SupplierFinder(automat.Automat):
    """
    This class implements all the functionality of the ``supplier_finder()``
    state machine.
    """

    timers = {
        'timer-30sec': (30.0, ['SERVICE?']),
        'timer-10sec': (10.0, ['ACK?']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self.target_idurl = None

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'start' and not self.isSomeCandidatesListed(*args, **kwargs):
                self.state = 'RANDOM_USER'
                self.Attempts = 0
                self.doInit(*args, **kwargs)
                self.doDHTFindRandomUser(*args, **kwargs)
            elif event == 'start' and self.isSomeCandidatesListed(*args, **kwargs):
                self.state = 'ACK?'
                self.doInit(*args, **kwargs)
                self.doPopCandidate(*args, **kwargs)
                self.Attempts = 1
                self.doSendMyIdentity(*args, **kwargs)
        #---RANDOM_USER---
        elif self.state == 'RANDOM_USER':
            if event == 'users-not-found':
                self.state = 'FAILED'
                self.doDestroyMe(*args, **kwargs)
                self.doReportFailed(*args, **kwargs)
            elif event == 'found-one-user':
                self.state = 'ACK?'
                self.doCleanPrevUser(*args, **kwargs)
                self.doRememberUser(*args, **kwargs)
                self.Attempts += 1
                self.doSendMyIdentity(*args, **kwargs)
        #---ACK?---
        elif self.state == 'ACK?':
            if self.Attempts == 5 and (event == 'timer-10sec' or event == 'ping-failed'):
                self.state = 'FAILED'
                self.doDestroyMe(*args, **kwargs)
                self.doReportFailed(*args, **kwargs)
            elif (event == 'ping-failed' or event == 'timer-10sec') and self.Attempts < 5:
                self.state = 'RANDOM_USER'
                self.doDHTFindRandomUser(*args, **kwargs)
            elif event == 'ack-received':
                self.state = 'SERVICE?'
                self.doSupplierConnect(*args, **kwargs)
        #---SERVICE?---
        elif self.state == 'SERVICE?':
            if event == 'supplier-connected':
                self.state = 'DONE'
                self.doDestroyMe(*args, **kwargs)
                self.doReportDone(*args, **kwargs)
            elif self.Attempts < 5 and event == 'supplier-not-connected':
                self.state = 'RANDOM_USER'
                self.doDHTFindRandomUser(*args, **kwargs)
            elif event == 'timer-30sec' and self.Attempts < 5:
                self.state = 'RANDOM_USER'
                self.doDHTFindRandomUser(*args, **kwargs)
            elif self.Attempts == 5 and (event == 'timer-30sec' or event == 'supplier-not-connected'):
                self.state = 'FAILED'
                self.doDestroyMe(*args, **kwargs)
                self.doReportFailed(*args, **kwargs)
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        #---DONE---
        elif self.state == 'DONE':
            pass
        return None

    def isSomeCandidatesListed(self, *args, **kwargs):
        """
        Condition method.
        """
        global _SuppliersToHire
        available = []
        for idurl in _SuppliersToHire:
            if id_url.is_not_in(idurl, contactsdb.suppliers(), as_field=False):
                available.append(idurl)
        return len(available) > 0

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.family_position = kwargs.get('family_position', None)
        self.ecc_map = kwargs.get('ecc_map')
        self.family_snapshot = kwargs.get('family_snapshot')

    def doSendMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        d = online_status.ping(
            idurl=self.target_idurl,
            channel='supplier_finder',
            keep_alive=False,
        )
        d.addCallback(lambda ok: self.automat('ack-received', ok))
        d.addErrback(lambda err: self.automat('ping-failed'))

    def doSupplierConnect(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.customer import supplier_connector
        from bitdust.customer import fire_hire
        from bitdust.raid import eccmap
        position = self.family_position
        if position is None or position == -1:
            lg.warn('position for new supplier is unknown, will "guess"')
            current_suppliers = list(contactsdb.suppliers())
            for i in range(len(current_suppliers)):
                supplier_idurl = current_suppliers[i].to_bin()
                if not supplier_idurl:
                    position = i
                    break
                if id_url.is_in(supplier_idurl, fire_hire.A().dismiss_list, as_field=False):
                    position = i
                    break
        sc = supplier_connector.by_idurl(self.target_idurl)
        if not sc:
            sc = supplier_connector.create(
                supplier_idurl=self.target_idurl,
                customer_idurl=my_id.getIDURL(),
            )
        sc.set_callback('supplier_finder', self._supplier_connector_state)
        sc.automat(
            'connect',
            family_position=position,
            ecc_map=(self.ecc_map or eccmap.Current().name),
            family_snapshot=self.family_snapshot,
        )

    def doDHTFindRandomUser(self, *args, **kwargs):
        """
        Action method.
        """
        tsk = lookup.random_supplier(ignore_idurls=contactsdb.suppliers())
        tsk.result_defer.addCallback(self._nodes_lookup_finished)
        tsk.result_defer.addErrback(lambda err: self.automat('users-not-found'))

    def doCleanPrevUser(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.customer import supplier_connector
        if id_url.is_cached(self.target_idurl):
            sc = supplier_connector.by_idurl(self.target_idurl)
            if sc:
                sc.remove_callback('supplier_finder', self._supplier_connector_state)
        self.target_idurl = None

    def doRememberUser(self, *args, **kwargs):
        """
        Action method.
        """
        self.target_idurl = id_url.field(args[0])
        if _Debug:
            lg.args(_DebugLevel, target_idurl=self.target_idurl)

    def doPopCandidate(self, *args, **kwargs):
        """
        Action method.
        """
        global _SuppliersToHire
        for idurl in _SuppliersToHire:
            if id_url.is_not_in(idurl, contactsdb.suppliers(), as_field=False):
                self.target_idurl = id_url.field(idurl)
                _SuppliersToHire.remove(idurl)
                break
        lg.info('populate supplier %r from "hire" list, %d more in the list' % (self.target_idurl, len(_SuppliersToHire)))

    def doReportDone(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.customer import fire_hire
        fire_hire.A('supplier-connected', *args, **kwargs)

    def doReportFailed(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.customer import fire_hire
        fire_hire.A('search-failed', *args, **kwargs)

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        from bitdust.customer import supplier_connector
        global _SupplierFinder
        del _SupplierFinder
        _SupplierFinder = None
        if self.target_idurl:
            sc = supplier_connector.by_idurl(self.target_idurl)
            if sc:
                sc.remove_callback('supplier_finder', self._supplier_connector_state)
            self.target_idurl = None
        self.destroy()

    #------------------------------------------------------------------------------

    def _nodes_lookup_finished(self, idurls):
        if _Debug:
            lg.out(_DebugLevel, 'supplier_finder._nodes_lookup_finished : %r' % idurls)
        if not idurls:
            lg.warn('no available nodes found via DHT lookup')
            self.automat('users-not-found')
            return
        found_idurl = None
        for idurl in idurls:
            #             if id_url.is_in(idurl, contactsdb.suppliers(), as_field=True):
            #                 if _Debug:
            #                     lg.out('    skip %r because already my supplier' % idurl)
            #                 continue
            ident = identitycache.FromCache(idurl)
            remoteprotos = set(ident.getProtoOrder())
            myprotos = set(my_id.getLocalIdentity().getProtoOrder())
            if not len(myprotos.intersection(remoteprotos)):
                if _Debug:
                    lg.out(_DebugLevel, '    skip %r because no matching protocols exists' % idurl)
                continue
            found_idurl = idurl
            break
        if not found_idurl:
            lg.warn('found some nodes via DHT lookup, but none of them is available')
            self.automat('users-not-found')
            return
        if _Debug:
            lg.out(_DebugLevel, '    selected %r and will request supplier service' % found_idurl)
        self.automat('found-one-user', found_idurl)

    def _supplier_connector_state(self, supplier_idurl, newstate, **kwargs):
        if id_url.field(supplier_idurl) != self.target_idurl:
            return
        if newstate in ['DISCONNECTED', 'NO_SERVICE']:
            self.automat('supplier-not-connected')
            return
        if newstate != 'CONNECTED':
            return
        if contactsdb.is_supplier(self.target_idurl):
            return
        family_position = kwargs.get('family_position', None)
        ecc_map = kwargs.get('ecc_map')
        self.automat('supplier-connected', self.target_idurl, family_position=family_position, ecc_map=ecc_map, family_snapshot=self.family_snapshot)



#!/usr/bin/env python
# ppserver.py
#
# Parallel Python Software: http://www.parallelpython.com
# Copyright (c) 2005-2009, Vitalii Vanovschi
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the author nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
"""
Parallel Python Software, Network Server.

http://www.parallelpython.com - updates, documentation, examples and support
forums
"""

from __future__ import absolute_import
from __future__ import print_function
from six import text_type as str
from six.moves import range
import six.moves._thread

copyright = 'Copyright (c) 2005-2009 Vitalii Vanovschi. All rights reserved'
version = '1.5.7'

import logging
import getopt
import sys
import socket
import random
import string
import time
import os

from . import pptransport
from . import ppauto

from pp import Server  # @UnresolvedImport

# compartibility with Python 2.6
try:
    import hashlib
    sha_new = hashlib.sha1
except ImportError:
    import sha
    sha_new = sha.new


class _NetworkServer(Server):
    """
    Network Server Class.
    """
    def __init__(self, ncpus='autodetect', interface='0.0.0.0', broadcast='255.255.255.255', port=None, secret=None, timeout=None, loglevel=logging.WARNING, restart=False, proto=0):
        Server.__init__(self, ncpus, secret=secret, loglevel=loglevel, restart=restart, proto=proto)
        self.host = interface
        self.bcast = broadcast
        if port is not None:
            self.port = port
        else:
            self.port = self.default_port
        self.timeout = timeout
        self.ncon = 0
        self.last_con_time = time.time()
        self.ncon_lock = six.moves._thread.allocate_lock()

        logging.debug('Strarting network server interface=%s port=%i' % (self.host, self.port))
        if self.timeout is not None:
            logging.debug('ppserver will exit in %i seconds if no '
                          'connections with clients exist' % (self.timeout), )
            six.moves._thread.start_new_thread(self.check_timeout, ())

    def ncon_add(self, val):
        """
        Keeps track of the number of connections and time of the last one.
        """
        self.ncon_lock.acquire()
        self.ncon += val
        self.last_con_time = time.time()
        self.ncon_lock.release()

    def check_timeout(self):
        """
        Checks if timeout happened and shutdowns server if it did.
        """
        while True:
            if self.ncon == 0:
                idle_time = time.time() - self.last_con_time
                if idle_time < self.timeout:
                    time.sleep(self.timeout - idle_time)
                else:
                    logging.debug(
                        'exiting ppserver due to timeout (no client'
                        ' connections in last %i sec)',
                        self.timeout,
                    )
                    os._exit(0)
            else:
                time.sleep(self.timeout)

    def listen(self):
        """
        Initiates listenting to incoming connections.
        """
        try:
            ssocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # following allows ppserver to restart faster on the same port
            ssocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            ssocket.bind((self.host, self.port))
            ssocket.listen(5)
        except socket.error:
            logging.error('Cannot create socket with port ' + str(self.port) + ' (port is already in use)')

        try:
            while True:
                # accept connections from outside
                (csocket, address) = ssocket.accept()
                # now do something with the clientsocket
                # in this case, we'll pretend this is a threaded server
                six.moves._thread.start_new_thread(self.crun, (csocket, ))
        except:
            logging.debug('Closing server socket')
            ssocket.close()

    def crun(self, csocket):
        """
        Authenticates client and handles its jobs.
        """
        mysocket = pptransport.CSocketTransport(csocket)
        # send PP version
        mysocket.send(version)
        # generate a random string
        srandom = ''.join([random.choice(string.ascii_letters) for i in range(16)])
        mysocket.send(srandom)
        answer = sha_new(srandom + self.secret).hexdigest()
        cleintanswer = mysocket.receive()
        if answer != cleintanswer:
            logging.warning('Authentification failed, client host=%s, port=%i' % csocket.getpeername())
            mysocket.send('FAILED')
            csocket.close()
            return
        else:
            mysocket.send('OK')

        ctype = mysocket.receive()
        logging.debug('Control message received: ' + ctype)
        self.ncon_add(1)
        try:
            if ctype == 'STAT':
                # reset time at each new connection
                self.get_stats()['local'].time = 0.0
                mysocket.send(str(self.get_ncpus()))
                while True:
                    mysocket.receive()
                    mysocket.send(str(self.get_stats()['local'].time))
            elif ctype == 'EXEC':
                while True:
                    sfunc = mysocket.receive()
                    sargs = mysocket.receive()
                    fun = self.insert(sfunc, sargs)
                    sresult = fun(True)
                    mysocket.send(sresult)
        except:
            logging.debug('Closing client socket')
            csocket.close()
            self.ncon_add(-1)

    def broadcast(self):
        """
        Initiaates auto-discovery mechanism.
        """
        discover = ppauto.Discover(self)
        six.moves._thread.start_new_thread(
            discover.run,
            ((self.host, self.port), (self.bcast, self.port)),
        )


def parse_config(file_loc):
    """
    Parses a config file in a very forgiving way.
    """
    # If we don't have configobj installed then let the user know and exit
    try:
        from configobj import ConfigObj
    except ImportError as ie:
        print(
            'ERROR: You must have configobj installed to use \
configuration files. You can still use command line switches.',
            file=sys.stderr,
        )
        sys.exit(1)

    if not os.access(file_loc, os.F_OK):
        print('ERROR: Can not access %s.' % arg, file=sys.stderr)
        sys.exit(1)

    # Load the configuration file
    config = ConfigObj(file_loc)
    # try each config item and use the result if it exists. If it doesn't
    # then simply pass and move along
    try:
        args['secret'] = config['general'].get('secret')
    except:
        pass

    try:
        autodiscovery = config['network'].as_bool('autodiscovery')
    except:
        pass

    try:
        args['interface'] = config['network'].get('interface', default='0.0.0.0')
    except:
        pass

    try:
        args['broadcast'] = config['network'].get('broadcast')
    except:
        pass

    try:
        args['port'] = config['network'].as_int('port')
    except:
        pass

    try:
        args['loglevel'] = config['general'].as_bool('debug')
    except:
        pass

    try:
        args['ncpus'] = config['general'].as_int('workers')
    except:
        pass

    try:
        args['proto'] = config['general'].as_int('proto')
    except:
        pass

    try:
        args['restart'] = config['general'].as_bool('restart')
    except:
        pass

    try:
        args['timeout'] = config['network'].as_int('timeout')
    except:
        pass
    # Return a tuple of the args dict and autodiscovery variable
    return args, autodiscovery


def print_usage():
    """
    Prints help.
    """
    print('Parallel Python Network Server (pp-' + version + ')')
    print(
        'Usage: ppserver.py [-hdar] [-n proto] [-c config_path]'\
        ' [-i interface] [-b broadcast] [-p port] [-w nworkers]'\
        ' [-s secret] [-t seconds]',
    )
    print()
    print('Options: ')
    print('-h                 : this help message')
    print('-d                 : debug')
    print('-a                 : enable auto-discovery service')
    print(
        '-r                 : restart worker process after each'\
        ' task completion',
    )
    print('-n proto           : protocol number for pickle module')
    print('-c path            : path to config file')
    print('-i interface       : interface to listen')
    print('-b broadcast       : broadcast address for auto-discovery service')
    print('-p port            : port to listen')
    print('-w nworkers        : number of workers to start')
    print('-s secret          : secret for authentication')
    print(
        '-t seconds         : timeout to exit if no connections with '\
        'clients exist',
    )
    print()
    print('Due to the security concerns always use a non-trivial secret key.')
    print('Secret key set by -s switch will override secret key assigned by')
    print('pp_secret variable in .pythonrc.py')
    print()
    print('Please visit http://www.parallelpython.com for extended up-to-date')
    print('documentation, examples and support forums')


if __name__ == '__main__':
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'hdarn:c:b:i:p:w:s:t:', ['help'])
    except getopt.GetoptError:
        print_usage()
        sys.exit(1)

    args = {}
    autodiscovery = False

    for opt, arg in opts:
        if opt in ('-h', '--help'):
            print_usage()
            sys.exit()
        elif opt == '-c':
            args, autodiscovery = parse_config(arg)
        elif opt == '-d':
            args['loglevel'] = logging.DEBUG
        elif opt == '-i':
            args['interface'] = arg
        elif opt == '-s':
            args['secret'] = arg
        elif opt == '-p':
            args['port'] = int(arg)
        elif opt == '-w':
            args['ncpus'] = int(arg)
        elif opt == '-a':
            autodiscovery = True
        elif opt == '-r':
            args['restart'] = True
        elif opt == '-b':
            args['broadcast'] = arg
        elif opt == '-n':
            args['proto'] = int(arg)
        elif opt == '-t':
            args['timeout'] = int(arg)

    server = _NetworkServer(**args)
    if autodiscovery:
        server.broadcast()
    server.listen()
    # have to destroy it here explicitelly otherwise an exception
    # comes out in Python 2.4
    del server

# Parallel Python Software: http://www.parallelpython.com



#!/usr/bin/env python
# id_registrator.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (id_registrator.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: id_registrator.

.. role:: red


Creates new identities to be able to access BitDust network.
Another node in the network can start ID server and store a copy of your identity.
Then third node can access that ID server and download your identity.

Normally, software will do all stuff for you and just pick few random ID servers,
ping them and register your identity.

But you can decide which ID servers you prefer and modify your "known" ID servers:

    bitdust set services/identity-propagate/known-servers first-server.com:80,second-host.net:8080


Your global IDURL is formed based on your nickname and DNS name (or IP address) of the first ID server.

If one of your ID servers is down, you can find a fresh one and "propagate" your identity there
and then remove dead ID server from the list of your sources: "identity migration" (not implemented yet).
This process will be automated and network identification will become much more reliable.

To be able to test locally you can start your own local ID server:

    bitdust set services/identity-server/host localhost
    ~/.bitdust/venv/bin/python userid/id_server.py 8080


Modify your "known" ID servers:

    bitdust set services/identity-propagate/known-servers localhost:8080


Modify your settings to use only one ID server:

    bitdust set services/identity-propagate/min-servers 1
    bitdust set services/identity-propagate/max-servers 1


Then you can create a "local" identity:

    ~/.bitdust/venv/bin/python userid/id_registrator.py my_nickname_here localhost


EVENTS:
    * :red:`id-exist`
    * :red:`id-not-exist`
    * :red:`id-server-failed`
    * :red:`id-server-response`
    * :red:`local-ip-detected`
    * :red:`my-id-exist`
    * :red:`my-id-failed`
    * :red:`my-id-sent`
    * :red:`start`
    * :red:`stun-failed`
    * :red:`stun-success`
    * :red:`timer-15sec`
    * :red:`timer-2min`
    * :red:`timer-30sec`
    * :red:`timer-5sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import sys
import random

from twisted.internet.defer import DeferredList
from six.moves import range

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.system import bpio

from bitdust.lib import nameurl
from bitdust.lib import net_misc
from bitdust.lib import misc
from bitdust.lib import strng

from bitdust.main import settings
from bitdust.main import config

from bitdust.stun import stun_client

from bitdust.crypt import key

from bitdust.userid import my_id
from bitdust.userid import identity
from bitdust.userid import id_url
from bitdust.userid import known_servers

#------------------------------------------------------------------------------

_IdRegistrator = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _IdRegistrator
    if _IdRegistrator is None:
        # set automat name and starting state here
        _IdRegistrator = IdRegistrator(
            name='id_registrator',
            state='AT_STARTUP',
            debug_level=2,
            log_transitions=True,
            log_events=True,
            publish_events=True,
        )
    if event is not None:
        _IdRegistrator.automat(event, *args, **kwargs)
    return _IdRegistrator


class IdRegistrator(automat.Automat):
    """
    This class implements all the functionality of the ``id_registrator()``
    state machine.
    """

    timers = {
        'timer-2min': (120, ['SEND_ID']),
        'timer-30sec': (30.0, ['NAME_FREE?']),
        'timer-5sec': (5.0, ['REQUEST_ID']),
        'timer-15sec': (15.0, ['REQUEST_ID']),
    }

    MESSAGES = {
        'MSG_0': ['ping identity servers...'],
        'MSG_1': ['checking the availability of a user name'],
        'MSG_2': ['checking network configuration'],
        'MSG_3': ['detecting external IP...'],
        'MSG_4': ['registering on ID servers...'],
        'MSG_5': ['verifying my identity'],
        'MSG_6': ['new user %(login)s registered successfully!', 'green'],
        'MSG_7': ['ID servers not responding', 'red'],
        'MSG_8': ['name %(login)s already taken', 'red'],
        'MSG_9': ['network connection error', 'red'],
        'MSG_10': ['connection error while sending my identity', 'red'],
        'MSG_11': ['identity verification failed', 'red'],
        'MSG_12': ['time out requesting from identity servers', 'red'],
        'MSG_13': ['time out sending to identity servers', 'red'],
        'MSG_14': ['generating your Private Key'],
    }

    def msg(self, msgid, *args, **kwargs):
        msg = self.MESSAGES.get(msgid, ['', 'black'])
        text = msg[0] % {
            'login': strng.to_text(bpio.ReadTextFile(settings.UserNameFilename())),
            'externalip': strng.to_text(misc.readExternalIP()),
            'localip': strng.to_text(bpio.ReadTextFile(settings.LocalIPFilename())),
        }
        color = 'black'
        if len(msg) == 2:
            color = msg[1]
        return text, color

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self.known_servers = {}  # host : (web port, tcp port)
        self.preferred_servers = []
        self.min_servers = 1
        self.max_servers = 10
        self.discovered_servers = []
        self.good_servers = []
        self.registrations = []
        self.free_idurls = []
        self.new_identity = None
        self.last_message = ''

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        This method intended to catch the moment when automat's state was changed.
        """
        from bitdust.main import installer
        installer.A('id_registrator.state', newstate)

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'start':
                self.state = 'ID_SERVERS?'
                self.doSaveMyName(*args, **kwargs)
                self.doSelectRandomServers(*args, **kwargs)
                self.doPingServers(*args, **kwargs)
                self.doPrint(self.msg('MSG_0', *args, **kwargs))
        #---DONE---
        elif self.state == 'DONE':
            pass
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        #---ID_SERVERS?---
        elif self.state == 'ID_SERVERS?':
            if (event == 'id-server-response' or event == 'id-server-failed') and self.isAllTested(*args, **kwargs) and self.isSomeAlive(*args, **kwargs):
                self.state = 'NAME_FREE?'
                self.doRequestServers(*args, **kwargs)
                self.doPrint(self.msg('MSG_1', *args, **kwargs))
            elif (event == 'id-server-response' or event == 'id-server-failed') and self.isAllTested(*args, **kwargs) and not self.isSomeAlive(*args, **kwargs):
                self.state = 'FAILED'
                self.doPrint(self.msg('MSG_7', *args, **kwargs))
                self.doDestroyMe(*args, **kwargs)
        #---NAME_FREE?---
        elif self.state == 'NAME_FREE?':
            if event == 'id-not-exist' and self.isAllResponded(*args, **kwargs) and self.isFreeIDURLs(*args, **kwargs):
                self.state = 'LOCAL_IP'
                self.doDetectLocalIP(*args, **kwargs)
                self.doPrint(self.msg('MSG_2', *args, **kwargs))
            elif event == 'timer-30sec' or (event == 'id-exist' and self.isAllResponded(*args, **kwargs) and not self.isFreeIDURLs(*args, **kwargs)):
                self.state = 'FAILED'
                self.doPrint(self.msg('MSG_8', *args, **kwargs))
                self.doDestroyMe(*args, **kwargs)
        #---LOCAL_IP---
        elif self.state == 'LOCAL_IP':
            if event == 'local-ip-detected':
                self.state = 'EXTERNAL_IP'
                self.doStunExternalIP(*args, **kwargs)
                self.doPrint(self.msg('MSG_3', *args, **kwargs))
        #---EXTERNAL_IP---
        elif self.state == 'EXTERNAL_IP':
            if event == 'stun-success':
                self.state = 'SEND_ID'
                self.doPrint(self.msg('MSG_14', *args, **kwargs))
                self.doCreateMyIdentity(*args, **kwargs)
                self.doPrint(self.msg('MSG_4', *args, **kwargs))
                self.doSendMyIdentity(*args, **kwargs)
            elif event == 'stun-failed':
                self.state = 'FAILED'
                self.doPrint(self.msg('MSG_9', *args, **kwargs))
                self.doDestroyMe(*args, **kwargs)
        #---SEND_ID---
        elif self.state == 'SEND_ID':
            if event == 'my-id-sent':
                self.state = 'REQUEST_ID'
                self.doRequestMyIdentity(*args, **kwargs)
                self.doPrint(self.msg('MSG_5', *args, **kwargs))
            elif event == 'my-id-failed':
                self.state = 'FAILED'
                self.doPrint(self.msg('MSG_10', *args, **kwargs))
                self.doDestroyMe(*args, **kwargs)
            elif event == 'timer-2min':
                self.state = 'FAILED'
                self.doPrint(self.msg('MSG_13', *args, **kwargs))
                self.doDestroyMe(*args, **kwargs)
        #---REQUEST_ID---
        elif self.state == 'REQUEST_ID':
            if event == 'my-id-exist' and self.isMyIdentityValid(*args, **kwargs):
                self.state = 'DONE'
                self.doSaveMyIdentity(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
                self.doPrint(self.msg('MSG_6', *args, **kwargs))
            elif event == 'timer-5sec':
                self.doRequestMyIdentity(*args, **kwargs)
            elif event == 'my-id-exist' and not self.isMyIdentityValid(*args, **kwargs):
                self.state = 'FAILED'
                self.doPrint(self.msg('MSG_11', *args, **kwargs))
                self.doDestroyMe(*args, **kwargs)
            elif event == 'timer-15sec':
                self.state = 'FAILED'
                self.doPrint(self.msg('MSG_12', *args, **kwargs))
                self.doDestroyMe(*args, **kwargs)
        return None

    def isMyIdentityValid(self, *args, **kwargs):
        """
        Condition method.
        """
        id_from_server = identity.identity(xmlsrc=args[0])
        if not id_from_server.isCorrect():
            lg.warn('my identity is not correct')
            return False
        if not id_from_server.Valid():
            lg.warn('my identity is not valid')
            return False
        if self.new_identity.serialize() != id_from_server.serialize():
            # TODO: seems an issue here!
            lg.warn('my identity source is different than copy received from id server')
            return False
        return True

    def isSomeAlive(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.good_servers) > 0

    def isAllResponded(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.registrations) == 0

    def isAllTested(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.discovered_servers) == 0

    def isFreeIDURLs(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.free_idurls) > 0

    def doSaveMyName(self, *args, **kwargs):
        """
        Action method.
        """
        if args:
            login = args[0]['username']
            preferred_servers = args[0].get('preferred_servers', [])
        else:
            login = kwargs['username']
            preferred_servers = kwargs.get('preferred_servers', [])
        self.preferred_servers = [s.strip() for s in preferred_servers]
        if not self.known_servers:
            self.known_servers = known_servers.by_host()
        if not self.preferred_servers:
            try:
                for srv in strng.to_text(config.conf().getString('services/identity-propagate/preferred-servers')).split(','):
                    if srv.strip():
                        self.preferred_servers.append(srv.strip())
            except:
                pass
        self.min_servers = max(settings.MinimumIdentitySources(), config.conf().getInt('services/identity-propagate/min-servers') or settings.MinimumIdentitySources())
        self.max_servers = min(settings.MaximumIdentitySources(), config.conf().getInt('services/identity-propagate/max-servers') or settings.MaximumIdentitySources())
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator.doSaveMyName [%s]' % login)
            lg.out(_DebugLevel, '    known_servers=%s' % self.known_servers)
            lg.out(_DebugLevel, '    preferred_servers=%s' % self.preferred_servers)
            lg.out(_DebugLevel, '    min_servers=%s' % self.min_servers)
            lg.out(_DebugLevel, '    max_servers=%s' % self.max_servers)
        bpio.WriteTextFile(settings.UserNameFilename(), login)

    def doSelectRandomServers(self, *args, **kwargs):
        """
        Action method.
        TODO: we also can search available id servers in DHT network as well
        """
        if self.preferred_servers:
            self.discovered_servers.extend(self.preferred_servers)
        num_servers = random.randint(self.min_servers, self.max_servers)
        needed_servers = num_servers - len(self.discovered_servers)
        for _ in range(needed_servers):
            # take a list of all known servers (only host names)
            s = set(self.known_servers.keys())
            # exclude already discovered servers
            s.difference_update(self.discovered_servers)
            if len(s) > 0:
                # if found some known servers - just pick a random one
                self.discovered_servers.append(random.choice(list(s)))
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator.doSelectRandomServers %s' % str(self.discovered_servers))

    def doPingServers(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator.doPingServers    %d in list' % len(self.discovered_servers))

        def _cb(htmlsrc, id_server_host):
            if _Debug:
                lg.out(_DebugLevel, '            RESPONDED: %s' % id_server_host)
            if self.preferred_servers and id_server_host in self.preferred_servers:
                self.good_servers.insert(0, id_server_host)
            else:
                self.good_servers.append(id_server_host)
            self.discovered_servers.remove(id_server_host)
            self.automat('id-server-response', (id_server_host, htmlsrc))

        def _eb(err, id_server_host):
            if _Debug:
                lg.out(_DebugLevel, '               FAILED: %s : %s' % (id_server_host, err.getErrorMessage()))
            self.discovered_servers.remove(id_server_host)
            self.automat('id-server-failed', (id_server_host, err))

        for host in self.discovered_servers:
            webport, tcpport = known_servers.by_host().get(
                host,
                (settings.IdentityWebPort(), settings.IdentityServerPort()),
            )
            if webport == 80:
                webport = ''
            server_url = nameurl.UrlMake('http', strng.to_text(host), webport, '')
            if _Debug:
                lg.out(_DebugLevel, '               connecting to %s' % server_url)
            d = net_misc.getPageTwisted(server_url, timeout=7)
            d.addCallback(_cb, host)
            d.addErrback(_eb, host)

    def doRequestServers(self, *args, **kwargs):
        """
        Action method.
        """
        login = bpio.ReadTextFile(settings.UserNameFilename())

        def _cb(xmlsrc, idurl, host):
            if not xmlsrc:
                if self.preferred_servers and host in self.preferred_servers:
                    if self.preferred_servers[0] == host:
                        self.free_idurls.insert(0, idurl)
                    else:
                        self.free_idurls.insert(1, idurl)
                else:
                    self.free_idurls.append(idurl)
                self.registrations.remove(idurl)
                self.automat('id-not-exist', idurl)
            else:
                if _Debug:
                    lg.out(_DebugLevel, '                EXIST: %s' % idurl)
                self.registrations.remove(idurl)
                self.automat('id-exist', idurl)

        def _eb(err, idurl, host):
            if _Debug:
                lg.out(_DebugLevel, '            NOT EXIST: %s' % idurl)
            if self.preferred_servers and host in self.preferred_servers:
                if self.preferred_servers[0] == host:
                    self.free_idurls.insert(0, idurl)
                else:
                    self.free_idurls.insert(1, idurl)
            else:
                self.free_idurls.append(idurl)
            self.registrations.remove(idurl)
            self.automat('id-not-exist', idurl)

        for host in self.good_servers:
            webport, tcpport = known_servers.by_host().get(host, (settings.IdentityWebPort(), settings.IdentityServerPort()))
            if webport == 80:
                webport = ''
            idurl = nameurl.UrlMake('http', strng.to_text(host), webport, login + '.xml')
            if _Debug:
                lg.out(_DebugLevel, '    %s' % idurl)
            d = net_misc.getPageTwisted(idurl, timeout=10)
            d.addCallback(_cb, idurl, host)
            d.addErrback(_eb, idurl, host)
            self.registrations.append(idurl)
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator.doRequestServers login=%s registrations=%d' % (login, len(self.registrations)))

    def doDetectLocalIP(self, *args, **kwargs):
        """
        Action method.
        """
        localip = net_misc.getLocalIp()
        bpio.WriteTextFile(settings.LocalIPFilename(), localip)
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator.doDetectLocalIP [%s]' % localip)
        self.automat('local-ip-detected')

    def doStunExternalIP(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator.doStunExternalIP')
        if len(self.free_idurls) == 1:
            if self.free_idurls[0].count(b'localhost:') or self.free_idurls[0].count(b'127.0.0.1:'):
                # if you wish to create a local identity you do not need to stun external IP at all
                self.automat('stun-success', '127.0.0.1')
                return True

        def save(result):
            if _Debug:
                lg.out(_DebugLevel, '            external IP : %s' % result)
            if result['result'] != 'stun-success':
                self.automat('stun-failed')
                return
            ip = result['ip']
            bpio.WriteTextFile(settings.ExternalIPFilename(), ip)
            self.automat('stun-success', ip)

        rnd_udp_port = random.randint(
            settings.DefaultUDPPort(),
            settings.DefaultUDPPort() + 500,
        )
        rnd_dht_port = random.randint(
            settings.DefaultDHTPort(),
            settings.DefaultDHTPort() + 500,
        )
        d = stun_client.safe_stun(udp_port=rnd_udp_port, dht_port=rnd_dht_port)
        d.addCallback(save)
        d.addErrback(lambda _: self.automat('stun-failed'))
        return True

    def doCreateMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        self._create_new_identity()

    def doSendMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        # TODO: just to debug - skip sending to ID servers and go further
        # self.state = 'REQUEST_ID'
        # self.event('my-id-exist', self.new_identity.serialize())
        # return
        mycurrentidentity = None
        if my_id.isLocalIdentityReady():
            mycurrentidentity = my_id.getLocalIdentity()
        my_id.setLocalIdentity(self.new_identity)

        def _cb(x):
            my_id.setLocalIdentity(mycurrentidentity)
            self.automat('my-id-sent')

        def _eb(x):
            my_id.setLocalIdentity(mycurrentidentity)
            self.automat('my-id-failed')

        dl = self._send_new_identity()
        dl.addCallback(_cb)
        dl.addErrback(_eb)

    def doRequestMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator.doRequestMyIdentity')

        def _cb(src):
            # TODO: validate my identity and make sure other servers also stored
            self.automat('my-id-exist', src)

        def _eb(err):
            # TODO: this is actually can be an issue if one server failed but others are fine...
            self.automat('my-id-not-exist', err)

        for idurl in self.new_identity.getSources(as_originals=True):
            if _Debug:
                lg.out(_DebugLevel, '        %s' % idurl)
            d = net_misc.getPageTwisted(idurl, timeout=15)
            d.addCallback(_cb)
            d.addErrback(_eb)

    def doSaveMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator.doSaveMyIdentity %s' % self.new_identity)
        my_id.setLocalIdentity(self.new_identity)
        my_id.saveLocalIdentity()

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        self.executeStateChangedCallbacks(oldstate=None, newstate=self.state, event_string=None, args=args)
        self.destroy(dead_state=self.state)
        global _IdRegistrator
        _IdRegistrator = None

    def doPrint(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.main import installer
        installer.A().event('print', *args, **kwargs)
        self.last_message = args[0][0]
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator.doPrint: %s' % str(*args, **kwargs))

    def _create_new_identity(self):
        """
        Generate new Private key and new identity file.
        Reads some extra info from config files.
        """
        login = strng.to_bin(bpio.ReadTextFile(settings.UserNameFilename()))
        externalIP = strng.to_bin(misc.readExternalIP()) or b'127.0.0.1'
        if self.free_idurls[0].count(b'127.0.0.1'):
            externalIP = b'127.0.0.1'
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator._create_new_identity %s %s ' % (login, externalIP))

        my_id.forgetLocalIdentity()
        my_id.eraseLocalIdentity(do_backup=True)
        key.ForgetMyKey(erase_file=True, do_backup=True)
        key.InitMyKey()
        if not key.isMyKeyReady():
            key.GenerateNewKey()
        if _Debug:
            lg.out(_DebugLevel, '    my key is ready')
        ident = my_id.buildDefaultIdentity(name=login, ip=externalIP, idurls=self.free_idurls)
        my_identity_xmlsrc = ident.serialize(as_text=True)
        newfilename = settings.LocalIdentityFilename() + '.new'
        bpio.WriteTextFile(newfilename, my_identity_xmlsrc)
        try:
            id_url.identity_cached(ident)
        except:
            lg.exc()
        self.new_identity = ident
        if _Debug:
            lg.out(_DebugLevel, '    wrote %d bytes to %s' % (len(my_identity_xmlsrc), newfilename))

    def _send_new_identity(self):
        """
        Send created identity to the identity servers to register it.
        """
        if _Debug:
            lg.out(_DebugLevel, 'id_registrator._send_new_identity')
        dlist = []
        payload = self.new_identity.serialize(as_text=False)
        for idurl in self.new_identity.getSources(as_originals=True):
            self.free_idurls.remove(strng.to_bin(idurl))
            _, host, webport, filename = nameurl.UrlParse(idurl)
            url = net_misc.pack_address(
                (
                    host,
                    webport,
                ),
                proto='http',
            )
            dlist.append(net_misc.http_post_data(
                url=url,
                data=payload,
                connectTimeout=15,
            ))
            if _Debug:
                lg.args(_DebugLevel, url=url, filename=filename, size=len(payload))
        return DeferredList(dlist, fireOnOneCallback=True)


#------------------------------------------------------------------------------


def main():
    bpio.init()
    settings.init()
    lg.set_debug_level(20)
    from twisted.internet import reactor  # @UnresolvedImport
    if len(sys.argv) > 2:
        args = (sys.argv[1], sys.argv[2])
    else:
        args = (sys.argv[1])
    A().addStateChangedCallback(lambda *a: reactor.stop(), oldstate=None, newstate='DONE')  # @UndefinedVariable
    A().addStateChangedCallback(lambda *a: reactor.stop(), oldstate=None, newstate='FAILED')  # @UndefinedVariable
    reactor.callWhenRunning(A, 'start', args)  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable
    settings.shutdown()


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# bandwidth.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (bandwidth.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: bandwidth.

.. role:: red

Here are counted incoming and outgoing traffic.
Statistics are saved on the user's disk in the
folders /bandin and /bandout in the BitDust local data dir.
This is a daily stats - a single file for every day.

THIS MODULE IS CURRENTLY DISABLED

"""

from __future__ import absolute_import
import os
import time

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.userid import my_id

from bitdust.p2p import commands

from bitdust.system import bpio

from bitdust.lib import misc

from bitdust.main import settings

from bitdust.userid import id_url

#------------------------------------------------------------------------------

BandInDict = {}
BandOutDict = {}
CountTimeIn = 0
CountTimeOut = 0

#------------------------------------------------------------------------------


def init():
    """
    Got a filename for today, check if already exists, read today file, start
    counting.
    """
    global CountTimeIn
    global CountTimeOut
    lg.out(4, 'bandwidth.init')
    fin = filenameIN()
    fout = filenameOUT()
    if not os.path.isfile(fin):
        bpio.WriteTextFile(fin, '')
    if not os.path.isfile(fout):
        bpio.WriteTextFile(fout, '')
    read_bandwidthIN()
    read_bandwidthOUT()
    CountTimeIn = time.time()
    CountTimeOut = time.time()
    reactor.addSystemEventTrigger('before', 'shutdown', save)  # @UndefinedVariable


def shutdown():
    """

    """
    lg.out(4, 'bandwidth.shutdown')


def filenameIN(basename=None):
    """
    File name for incoming stats for today or other day.
    """
    if basename is None:
        basename = misc.gmtime2str('%d%m%y')
    return os.path.join(settings.BandwidthInDir(), basename)


def filenameOUT(basename=None):
    """
    File name for outgoing stats for today or other day.
    """
    if basename is None:
        basename = misc.gmtime2str('%d%m%y')
    return os.path.join(settings.BandwidthOutDir(), basename)


def save():
    """
    Writes today stats on disk.
    """
    lg.out(6, 'bandwidth.save')
    bpio._write_dict(filenameIN(), getBandwidthIN())
    bpio._write_dict(filenameOUT(), getBandwidthOUT())


def saveIN(basename=None):
    """
    Writes incoming stats for today on disk.
    """
    if basename is None:
        basename = misc.gmtime2str('%d%m%y')
    ret = os.path.isfile(filenameIN(basename))
    bpio._write_dict(filenameIN(basename), getBandwidthIN())
    if not ret:
        lg.out(4, 'bandwidth.saveIN to new file ' + basename)
    else:
        lg.out(22, 'bandwidth.saveIN to ' + basename)
    return ret


def saveOUT(basename=None):
    """
    Writes outgoing stats for today on disk.
    """
    if basename is None:
        basename = misc.gmtime2str('%d%m%y')
    ret = os.path.isfile(filenameOUT(basename))
    bpio._write_dict(filenameOUT(basename), getBandwidthOUT())
    if not ret:
        lg.out(4, 'bandwidth.saveOUT to new file ' + basename)
    else:
        lg.out(22, 'bandwidth.saveOUT to ' + basename)
    return ret


def read_bandwidthIN():
    """
    Reads today's incoming bandwidth stats from disk.
    """
    global BandInDict
    lg.out(6, 'bandwidth.read_bandwidthIN ')
    for idurl, bytesin in bpio._read_dict(filenameIN(), {}).items():
        BandInDict[idurl] = int(bytesin)


def read_bandwidthOUT():
    """
    Reads today's outgoing bandwidth stats from disk.
    """
    global BandOutDict
    lg.out(6, 'bandwidth.read_bandwidthOUT ')
    for idurl, bytesout in bpio._read_dict(filenameOUT(), {}).items():
        BandOutDict[idurl] = int(bytesout)


def clear_bandwidthIN():
    """
    Erase all incoming stats from memory.
    """
    global BandInDict
    lg.out(6, 'bandwidth.clear_bandwidthIN ')
    BandInDict.clear()


def clear_bandwidthOUT():
    """
    Erase all outgoing stats from memory.
    """
    global BandOutDict
    lg.out(6, 'bandwidth.clear_bandwidthOUT ')
    BandOutDict.clear()


def clear():
    """
    Erase all bandwidth stats from memory.
    """
    clear_bandwidthIN()
    clear_bandwidthOUT()


def getBandwidthIN():
    """
    Get current incoming bandwidth stats from memory.
    """
    global BandInDict
    return BandInDict


def getBandwidthOUT():
    """
    Get current outgoing bandwidth stats from memory.
    """
    global BandOutDict
    return BandOutDict


def isExistIN():
    """
    Check existence of today's incoming bandwidth file on disk.
    """
    return os.path.isfile(filenameIN())


def isExistOUT():
    """
    Check existence of today's outgoing bandwidth file on disk.
    """
    return os.path.isfile(filenameOUT())


def files2send():
    """
    Return a list of file names to be read and send later.

    Sent files are market with ".sent" extension and skipped here.
    """
    lg.out(6, 'bandwidth.files2send')
    listIN = []
    listOUT = []
    for filename in os.listdir(settings.BandwidthInDir()):
        # if we sent the file - skip it
        if filename.endswith('.sent'):
            continue
        # if filename is not a date - skip it
        if len(filename) != 6:
            continue
        # skip today bandwidth - it is still counting, right?
        if filename == misc.gmtime2str('%d%m%y'):
            continue
        filepath = os.path.join(settings.BandwidthInDir(), filename)
        # if filepath == filenameIN():
        #     continue
        listIN.append(filepath)
    for filename in os.listdir(settings.BandwidthOutDir()):
        if filename.endswith('.sent'):
            continue
        if len(filename) != 6:
            continue
        if filename == misc.gmtime2str('%d%m%y'):  # time.strftime('%d%m%y'):
            continue
        filepath = os.path.join(settings.BandwidthOutDir(), filename)
        # if filepath == filenameOUT():
        #     continue
        listOUT.append(filepath)
    lg.out(6, 'bandwidth.files2send listIN=%d listOUT=%d' % (len(listIN), len(listOUT)))
    for i in listIN:
        lg.out(6, '  ' + i)
    for i in listOUT:
        lg.out(6, '  ' + i)
    return listIN, listOUT


def IN(idurl, size):
    """
    Call this when need to count incoming bandwidth.

    ``size`` - how many incoming bytes received from user with ``idurl``.
    Typically called when incoming packet arrives.
    """
    global BandInDict
    global CountTimeIn
    if not isExistIN():
        currentFileName = time.strftime('%d%m%y', time.localtime(CountTimeIn))
        if currentFileName != misc.gmtime2str('%d%m%y'):
            saveIN(currentFileName)
            clear_bandwidthIN()
            CountTimeIn = time.time()
    currentV = int(BandInDict.get(idurl, 0))
    newV = currentV + size
    BandInDict[idurl] = newV
    curMB = int(currentV/(1024.0*1024.0))
    newMB = int(newV/(1024.0*1024.0))
    if curMB == 0 or curMB != newMB:
        saveIN()


def OUT(idurl, size):
    """
    Call this when need to count outgoing bandwidth.

    ``size`` - how many bytes sent to user with ``idurl``.
    Typically called when outgoing packet were sent.
    """
    global BandOutDict
    global CountTimeOut
    if not isExistOUT():
        currentFileName = time.strftime('%d%m%y', time.localtime(CountTimeOut))
        if currentFileName != misc.gmtime2str('%d%m%y'):
            saveOUT(currentFileName)
            clear_bandwidthOUT()
            CountTimeOut = time.time()
    currentV = int(BandOutDict.get(idurl, 0))
    newV = currentV + size
    BandOutDict[idurl] = newV
    curMB = int(currentV/(1024.0*1024.0))
    newMB = int(newV/(1024.0*1024.0))
    if curMB == 0 or curMB != newMB:
        saveOUT()


def INfile(newpacket, pkt_in, status, error_message):
    """
    Count incoming file from ``proto``://``host``, ``newpacket`` is already
    Unserialized.
    """
    if status != 'finished':
        return False
    if not id_url.is_cached(newpacket.OwnerID):
        return False
    packet_from = newpacket.OwnerID
    if packet_from == my_id.getIDURL() and newpacket.Command == commands.Data():
        # someone giving our data back
        packet_from = newpacket.RemoteID
    if pkt_in.size:
        IN(packet_from, pkt_in.size)
    # IN(packet_from, len(newpacket.Payload))
    return False


def OUTfile(pkt_out, item, status, size, error_message):
    """
    Count outgoing file to ``proto``://``host``, ``workitem`` is from sending
    queue.
    """
    if status != 'finished':
        return False
    if pkt_out.filesize and pkt_out.remote_idurl:
        OUT(pkt_out.remote_idurl, pkt_out.filesize)
    # OUT(workitem.remoteid, workitem.payloadsize)
    return False



#!/usr/bin/env python
# shared_access_donor.py
#
# Copyright (C) 2008 Veselin Penev, http://bitdust.io
#
# This file (shared_access_donor.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: shared_access_donor
.. role:: red

BitDust shared_access_donor() Automat

EVENTS:
    * :red:`ack`
    * :red:`all-suppliers-acked`
    * :red:`audit-ok`
    * :red:`blockchain-ok`
    * :red:`fail`
    * :red:`init`
    * :red:`list-files-ok`
    * :red:`priv-key-ok`
    * :red:`timer-10sec`
    * :red:`timer-15sec`
    * :red:`timer-2sec`
    * :red:`timer-30sec`
    * :red:`user-identity-cached`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import time

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import packetid
from bitdust.lib import serialization
from bitdust.lib import strng

from bitdust.main import events

from bitdust.contacts import identitycache
from bitdust.contacts import contactsdb

from bitdust.p2p import p2p_service
from bitdust.p2p import online_status
from bitdust.p2p import commands

from bitdust.crypt import key
from bitdust.crypt import my_keys
from bitdust.crypt import encrypted

from bitdust.userid import my_id
from bitdust.userid import global_id
from bitdust.userid import id_url

from bitdust.access import key_ring

from bitdust.storage import backup_fs

#------------------------------------------------------------------------------


class SharedAccessDonor(automat.Automat):
    """
    This class implements all the functionality of the ``shared_access_donor()`` state machine.
    """

    timers = {
        'timer-10sec': (10.0, ['CACHE']),
        'timer-15sec': (15.0, ['PUB_KEY', 'PING', 'AUDIT', 'LIST_FILES']),
        'timer-2sec': (2.0, ['PUB_KEY']),
        'timer-30sec': (30.0, ['PRIV_KEY']),
    }

    def __init__(self, debug_level=0, log_events=False, publish_events=False, **kwargs):
        """
        Create shared_access_donor() state machine.
        Use this method if you need to call Automat.__init__() in a special way.
        """
        super(SharedAccessDonor, self).__init__(name='shared_access_donor', state='AT_STARTUP', debug_level=debug_level or _DebugLevel, log_events=log_events or _Debug, publish_events=publish_events, **kwargs)

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of shared_access_donor() machine.
        """
        self.log_transitions = _Debug
        self.caching_deferred = None
        self.remote_idurl = None
        self.remote_identity = None
        self.ping_response = None
        self.key_id = None
        self.result_defer = None
        self.suppliers_responses = {}
        self.suppliers_acks = 0

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'CACHE'
                self.doInit(*args, **kwargs)
                self.doInsertInboxCallback(*args, **kwargs)
                self.doCacheRemoteIdentity(*args, **kwargs)
        #---PING---
        elif self.state == 'PING':
            if event == 'ack':
                self.state = 'BLOCKCHAIN'
                self.doBlockchainLookupVerifyUserPubKey(*args, **kwargs)
            elif event == 'fail' or event == 'timer-15sec':
                self.state = 'CLOSED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---CACHE---
        elif self.state == 'CACHE':
            if event == 'user-identity-cached':
                self.state = 'PING'
                self.doSendMyIdentityToUser(*args, **kwargs)
            elif event == 'fail' or event == 'timer-10sec':
                self.state = 'CLOSED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---BLOCKCHAIN---
        elif self.state == 'BLOCKCHAIN':
            if event == 'fail':
                self.state = 'CLOSED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'blockchain-ok':
                self.state = 'AUDIT'
                self.doAuditUserMasterKey(*args, **kwargs)
        #---LIST_FILES---
        elif self.state == 'LIST_FILES':
            if event == 'list-files-ok':
                self.state = 'CLOSED'
                self.doReportDone(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'fail' or event == 'timer-15sec':
                self.state = 'CLOSED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---AUDIT---
        elif self.state == 'AUDIT':
            if event == 'audit-ok':
                self.state = 'PUB_KEY'
                self.doSendPubKeyToSuppliers(*args, **kwargs)
            elif event == 'fail' or event == 'timer-15sec':
                self.state = 'CLOSED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---PUB_KEY---
        elif self.state == 'PUB_KEY':
            if event == 'ack':
                self.doCheckAllAcked(*args, **kwargs)
            elif event == 'fail' or (event == 'timer-15sec' and not self.isSomeSuppliersAcked(*args, **kwargs)):
                self.state = 'CLOSED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'all-suppliers-acked' or (event == 'timer-2sec' and self.isSomeSuppliersAcked(*args, **kwargs)):
                self.state = 'PRIV_KEY'
                self.doSendPrivKeyToUser(*args, **kwargs)
        #---PRIV_KEY---
        elif self.state == 'PRIV_KEY':
            if event == 'priv-key-ok':
                self.state = 'LIST_FILES'
                self.doSendMyListFiles(*args, **kwargs)
            elif event == 'fail' or event == 'timer-30sec':
                self.state = 'CLOSED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isSomeSuppliersAcked(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.suppliers_acks > 0

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.remote_idurl = id_url.field(kwargs['trusted_idurl'])
        self.key_id = strng.to_text(kwargs['key_id'])
        self.result_defer = kwargs.get('result_defer', None)

    def doInsertInboxCallback(self, *args, **kwargs):
        """
        Action method.
        """

    def doCacheRemoteIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        self.caching_deferred = identitycache.immediatelyCaching(self.remote_idurl)
        self.caching_deferred.addCallback(self._on_remote_identity_cached)
        self.caching_deferred.addErrback(lambda err: self.automat('fail', err) and None)

    def doSendMyIdentityToUser(self, *args, **kwargs):
        """
        Action method.
        """
        def _on_ack(response):
            self.ping_response = time.time()
            self.automat('ack', response)
            return response

        def _on_fail(err):
            self.automat('fail', err)
            return None

        d = online_status.ping(
            idurl=self.remote_idurl,
            channel='shared_access_donor',
            ping_retries=1,
            keep_alive=True,
        )
        d.addCallback(_on_ack)
        d.addErrback(_on_fail)

    def doBlockchainLookupVerifyUserPubKey(self, *args, **kwargs):
        """
        Action method.
        """
        # TODO:
        self.automat('blockchain-ok')

    def doAuditUserMasterKey(self, *args, **kwargs):
        """
        Action method.
        """
        master_key_id = my_keys.make_key_id(alias='master', creator_idurl=self.remote_idurl)
        d = key_ring.audit_private_key(master_key_id, self.remote_idurl)
        d.addCallback(lambda audit_result: (self.automat('audit-ok') if audit_result else self.automat('fail', Exception('remote user master key audit process failed')), ))
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='shared_access_donor.doAuditUserMasterKey')
        d.addErrback(lambda err: self.automat('fail', err))

    def doSendPubKeyToSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        if not my_keys.is_key_registered(self.key_id):
            self.automat('fail', Exception('key not found'))
            return
        self.suppliers_acks = 0
        for supplier_idurl in contactsdb.suppliers():
            if supplier_idurl:
                d = key_ring.share_key(self.key_id, supplier_idurl, include_private=False, include_signature=False)
                d.addCallback(self._on_supplier_pub_key_shared, supplier_idurl)
                d.addErrback(self._on_supplier_pub_key_failed, supplier_idurl)
                self.suppliers_responses[supplier_idurl] = d

    def doCheckAllAcked(self, *args, **kwargs):
        """
        Action method.
        """
        if len(self.suppliers_responses) == 0:
            self.automat('all-suppliers-acked')

    def doSendPrivKeyToUser(self, *args, **kwargs):
        """
        Action method.
        """
        d = key_ring.share_key(self.key_id, self.remote_idurl, include_private=True, include_signature=True)
        d.addCallback(self._on_user_priv_key_shared)
        d.addErrback(self._on_user_priv_key_failed)

    def doSendMyListFiles(self, *args, **kwargs):
        """
        Action method.
        """
        self.automat('list-files-ok')
        return
        # TODO: cleanup
        json_list_files = backup_fs.SerializeIndex(
            customer_idurl=global_id.glob2idurl(self.key_id),
            to_json=True,
            filter_cb=lambda path_id, path, info: True if strng.to_text(info.key_id) == strng.to_text(self.key_id) else False,
        )
        raw_list_files = serialization.DictToBytes(json_list_files, keys_to_text=True, values_to_text=True, encoding='utf-8')
        if _Debug:
            lg.out(_DebugLevel, 'shared_access_donor.doSendMyListFiles prepared list of files for %s :\n%s' % (self.remote_idurl, raw_list_files))
        block = encrypted.Block(
            CreatorID=my_id.getIDURL(),
            BackupID=self.key_id,
            Data=raw_list_files,
            SessionKey=key.NewSessionKey(session_key_type=key.SessionKeyType()),
            SessionKeyType=key.SessionKeyType(),
            EncryptKey=self.key_id,
        )
        encrypted_list_files = block.Serialize()
        packet_id = '%s:%s' % (
            self.key_id,
            packetid.UniqueID(),
        )
        p2p_service.SendFiles(
            idurl=self.remote_idurl,
            raw_list_files_info=encrypted_list_files,
            packet_id=packet_id,
            callbacks={
                commands.Ack(): lambda response, _: self.automat('list-files-ok', response),
                commands.Fail(): lambda response, _: self.automat('fail', Exception(str(response))),
                None: lambda pkt_out: self.automat('fail', Exception('timeout')),
            },
        )

    def doReportDone(self, *args, **kwargs):
        """
        Action method.
        """
        lg.info('share key [%s] with %r finished with success' % (self.key_id, self.remote_idurl))
        events.send('private-key-shared', data=dict(
            global_id=global_id.UrlToGlobalID(self.remote_idurl),
            remote_idurl=self.remote_idurl,
            key_id=self.key_id,
        ))
        if self.result_defer:
            self.result_defer.callback(True)

    def doReportFailed(self, event, *args, **kwargs):
        """
        Action method.
        """
        lg.warn('share key [%s] with %s failed: %s' % (self.key_id, self.remote_idurl, args))
        reason = 'key transfer failed with unknown reason'
        if args and args[0]:
            reason = args[0]
        else:
            if self.remote_identity is None:
                reason = 'remote identity caching failed',
            else:
                if self.ping_response is None:
                    reason = 'remote node not responding',
                else:
                    if self.suppliers_responses:
                        reason = 'connection timeout with my suppliers'
                    else:
                        if event.count('timer-'):
                            reason = 'key transfer failed because of network connection timeout'
        events.send('private-key-share-failed', data=dict(
            global_id=global_id.UrlToGlobalID(self.remote_idurl),
            remote_idurl=self.remote_idurl,
            key_id=self.key_id,
            reason=reason,
        ))
        if self.result_defer:
            self.result_defer.errback(Exception(reason))

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.remote_idurl = None
        self.remote_identity = None
        self.ping_response = None
        self.key_id = None
        self.result_defer = None
        self.suppliers_responses.clear()
        self.suppliers_acks = 0
        if self.caching_deferred:
            self.caching_deferred.cancel()
            self.caching_deferred = None
        self.destroy()

    def _on_remote_identity_cached(self, xmlsrc):
        self.caching_deferred = None
        self.remote_identity = contactsdb.get_contact_identity(self.remote_idurl)
        if self.remote_identity is None:
            self.automat('fail', Exception('remote id caching failed'))
        else:
            self.automat('user-identity-cached')

    def _on_supplier_pub_key_shared(self, response, supplier_idurl):
        self.suppliers_responses.pop(supplier_idurl, None)
        self.suppliers_acks += 1
        self.automat('ack', response)
        lg.warn('suppliers_acks=%d, suppliers_responses=%d' % (self.suppliers_acks, len(self.suppliers_responses)))
        return None

    def _on_supplier_pub_key_failed(self, err, supplier_idurl):
        lg.warn(err)
        self.suppliers_responses.pop(supplier_idurl, None)
        return None

    def _on_user_priv_key_shared(self, response):
        lg.info('your private key %s was sent to %s' % (self.key_id, self.remote_idurl))
        self.automat('priv-key-ok', response)
        return None

    def _on_user_priv_key_failed(self, err):
        lg.warn(err)
        self.automat('fail', Exception('private key delivery failed to remote node'))
        return None



"""
check_tx.py

A Demo script that takes a transaction id as input, and sends back a json with it's status
- unknown
- in mempool (shows timestamp, from, to, amount, first 50 chars of openfield)
- in ledger (shows timestamp, from, to, amount, first 50 chars of openfield as well as number of confirmations)

This script is supposed to be run in the node directory, it needs access to both ledger.db and mempool.db
"""

import sqlite3, sys, json

# Default ledger path
ledger_path = 'static/ledger.db'

# Default mempool path
mempool_path = 'mempool.db'


def list_to_tx(result):
    """
	Converts the query result into a dict
	"""
    result[4] = result[4][:50]
    keys = ['Timestamp', 'Address', 'Recipient', 'Amount', 'Openfield', 'Block']
    return dict(zip(keys, result))


def is_in_mempool(txid):
    """
	If txid is in mempool, sends back details of the tx
	"""
    mempool = sqlite3.connect(mempool_path)
    mempool.text_factory = str
    m = mempool.cursor()
    m.execute('SELECT timestamp, address, recipient, amount, openfield FROM transactions WHERE signature like ?;', (txid + '%', ))
    result = m.fetchone()
    if result:
        return (True, list_to_tx(list(result)))
    else:
        return (False, None)


def is_in_ledger(txid):
    """
	If txid is in ledger, sends back details of the tx and number of confirmations
	"""
    ledger = sqlite3.connect(ledger_path, timeout=1)
    ledger.text_factory = str
    m = ledger.cursor()
    m.execute('SELECT timestamp, address, recipient, amount, openfield, block_height FROM transactions WHERE signature like ?;', (txid + '%', ))
    result = m.fetchone()
    if result:
        m.execute('SELECT block_height FROM transactions ORDER BY block_height desc LIMIT 1')
        last = m.fetchone()
        return (True, list_to_tx(list(result)), last[0])
    else:
        return (False, None, None)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        txid = input('No argument detected, please insert command manually\n')
    else:
        txid = sys.argv[1]

    res = {'TxId': txid, 'Status': 'Unknown'}

    isit, details = is_in_mempool(txid)
    if isit:
        res['Status'] = 'Mempool'
        res.update(details)

    isit, details, lastblock = is_in_ledger(txid)
    if isit:
        res['Status'] = 'Confirmed'
        res.update(details)
        res['Confirmations'] = lastblock - res['Block']

    print(json.dumps(res))



#!/usr/bin/python
# certificate.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (certificate.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: certificate.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import datetime
import ipaddress
import uuid

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes
from cryptography.x509.oid import NameOID

#------------------------------------------------------------------------------


def generate_private_key(key_size=2048):
    """
    Create a private key
    """
    pkey = rsa.generate_private_key(public_exponent=65537, key_size=key_size, backend=default_backend())
    pkey_pem = pkey.private_bytes(
        encoding=serialization.Encoding.PEM,
        # format=serialization.PrivateFormat.PKCS8,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return pkey, pkey_pem


def load_private_key(pkey_pem):
    """
    Load private key from PEM text
    """
    pkey = serialization.load_pem_private_key(
        data=pkey_pem,
        password=None,
        backend=default_backend(),
    )
    return pkey


def load_certificate(cert_pem):
    """
    Load private key from PEM text
    """
    pub_key = x509.load_pem_x509_certificate(
        data=cert_pem,
        backend=default_backend(),
    )
    return pub_key


def generate_self_signed_cert(hostname, server_key, ip_addresses=None):
    """
    Generates self signed certificate for a hostname, and optional IP addresses.
    """
    name = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, hostname)])
    alt_names = [x509.DNSName(hostname)]
    if ip_addresses:
        for addr in ip_addresses:
            alt_names.append(x509.DNSName(addr))
            alt_names.append(x509.IPAddress(ipaddress.ip_address(addr)))

    ca_crt = x509.CertificateBuilder().subject_name(name).issuer_name(name).public_key(server_key.public_key()).serial_number(uuid.uuid4().int, ).not_valid_before(
        datetime.datetime.utcnow()
    ).not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365*100)).add_extension(
        extension=x509.KeyUsage(digital_signature=True, key_encipherment=True, key_cert_sign=True, crl_sign=True, content_commitment=True, data_encipherment=False, key_agreement=False, encipher_only=False, decipher_only=False),
        critical=False,
    ).add_extension(
        extension=x509.BasicConstraints(ca=True, path_length=0),
        critical=False,
    ).add_extension(
        extension=x509.SubjectKeyIdentifier.from_public_key(server_key.public_key()),
        critical=False,
    ).add_extension(
        extension=x509.AuthorityKeyIdentifier.from_issuer_public_key(server_key.public_key()),
        critical=False,
    ).add_extension(
        extension=x509.SubjectAlternativeName(alt_names),
        critical=False,
    ).sign(
        private_key=server_key,
        algorithm=hashes.SHA256(),
        backend=default_backend(),
    )
    ca_cert_pem = ca_crt.public_bytes(encoding=serialization.Encoding.PEM)

    return ca_cert_pem


def generate_csr_client_cert(hostname, server_ca_cert, server_key, client_key):
    """
    Generate client-side certificate with CSR.
    """
    csr = x509.CertificateSigningRequestBuilder().subject_name(x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, hostname),
    ]), ).add_extension(
        x509.SubjectAlternativeName([
            x509.DNSName(hostname),
        ]),
        critical=False,
    ).sign(client_key, hashes.SHA256(), default_backend())
    crt = x509.CertificateBuilder().subject_name(csr.subject).issuer_name(server_ca_cert.subject).public_key(csr.public_key()).serial_number(uuid.uuid4().int, ).not_valid_before(
        datetime.datetime.utcnow()
    ).not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365*100)).add_extension(
        extension=x509.KeyUsage(digital_signature=True, key_encipherment=True, content_commitment=True, data_encipherment=False, key_agreement=False, encipher_only=False, decipher_only=False, key_cert_sign=False, crl_sign=False),
        critical=True,
    ).add_extension(
        extension=x509.BasicConstraints(ca=False, path_length=None),
        critical=True,
    ).add_extension(
        extension=x509.AuthorityKeyIdentifier.from_issuer_public_key(server_key.public_key()),
        critical=False,
    ).sign(
        private_key=server_key,
        algorithm=hashes.SHA256(),
        backend=default_backend(),
    )
    crt_pem = crt.public_bytes(serialization.Encoding.PEM)
    return crt_pem



#!/usr/bin/python
# misc.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (misc.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: misc.

A set of different methods across the code.

TODO:
    Really need to do some refactoring here - too many things in one place.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
import six.moves.urllib.parse  # @UnresolvedImport
from six.moves import range  # @UnresolvedImport

#------------------------------------------------------------------------------

import os
import re
import sys
import time
import math
import random
import base64
import string
import locale
import hashlib
import tempfile
import textwrap
import functools
import subprocess

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import local_fs

from bitdust.main import settings

from bitdust.lib import packetid
from bitdust.lib import strng

#------------------------------------------------------------------------------

# if we come up with more valid transports,
# we'll need to add them here
# in the case we don't have any valid transports,
# we use this array as the default
# order is important!
# this is default order to be used for new users
# more stable transports must be higher
_AttenuationFactor = 2.0

#-------------------------------------------------------------------------------


def init():
    """
    Will be called in main thread at start up.

    Can put here some minor things if needed.
    """
    lg.out(4, 'misc.init')


#------------------------------------------------------------------------------


def readLocalIP():
    """
    Read local IP stored in the file `~/.bitdust/[network name]/metadata/localip`.
    """
    return bpio.ReadTextFile(settings.LocalIPFilename()).strip()


def readExternalIP():
    """
    Read external IP stored in the file `~/.bitdust/[network name]/metadata/externalip`.
    """
    return bpio.ReadTextFile(settings.ExternalIPFilename()).strip()


def readSupplierData(supplier_idurl, filename, customer_idurl):
    """
    Read a file from [BitDust data dir]/suppliers/[IDURL] folder.

    The file names right now is ['connected', 'disconnected', 'listfiles'].
    """
    path = settings.SupplierPath(supplier_idurl, customer_idurl, filename)
    if not os.path.isfile(path):
        return ''
    return bpio.ReadTextFile(path).strip()


def writeSupplierData(supplier_idurl, filename, data, customer_idurl):
    """
    Writes to a config file for given supplier.
    """
    dirPath = settings.SupplierPath(supplier_idurl, customer_idurl)
    if not os.path.isdir(dirPath):
        os.makedirs(dirPath)
    path = settings.SupplierPath(supplier_idurl, customer_idurl, filename)
    return bpio.WriteTextFile(path, data)


#-------------------------------------------------------------------------------


def cmp(a, b):
    return (a > b) - (a < b)


def NewBackupID(time_st=None):
    """
    BackupID is just a string representing time and date.

    Symbol "F" is placed at the start to identify that this is a FULL
    backup. We have a plans to provide INCREMENTAL backups also.
    """
    if time_st is None:
        time_st = time.localtime()
    ampm = time.strftime('%p', time_st)
    if not ampm:
        lg.warn('time.strftime() returns empty string')
        ampm = 'AM' if time.time() % 86400 < 43200 else 'PM'
    result = 'F' + time.strftime('%Y%m%d%I%M%S', time_st) + ampm
    return result


def TimeStructFromVersion(backupID):
    try:
        if backupID.endswith('AM') or backupID.endswith('PM'):
            ampm = backupID[-2:]
            st_time = list(time.strptime(backupID[1:-2], '%Y%m%d%I%M%S'))
        else:
            i = backupID.rfind('M')
            ampm = backupID[i - 1:i + 1]
            st_time = list(time.strptime(backupID[1:i - 1], '%Y%m%d%I%M%S'))
        if ampm == 'PM':
            st_time[3] += 12
        return tuple(st_time)
    except:
        lg.exc()
        return None


def TimeFromBackupID(backupID):
    """
    Reverse method - return a date and time from given BackupID with ``time.mktime()``.
    """
    try:
        return time.mktime(TimeStructFromVersion(backupID))
    except:
        lg.exc('backupID=%r' % backupID)
        return None


def modified_version(a):
    """
    Next functions are to come up with a sorted list of backup ids (dealing
    with AM/PM).

    This method make a number for given BackupID - used to compare two BackupID's.
    """
    try:
        if a.endswith('AM') or a.endswith('PM'):
            int_a = int(a[1:-2])
            int_b = 0
        else:
            i = a.rfind('M')
            int_a = int(a[1:i - 1])
            int_b = int(a[i + 1:])
    except:
        lg.exc()
        return -1
    hour = a[-8:-6]
    if a.endswith('PM') and hour != '12':
        int_a += 120000
    elif a.endswith('AM') and hour == '12':
        int_a -= 120000
    return int_a + int_b


def version_compare(version1, version2):
    """
    Compare two BackupID's, I start using another term for BackupID not so long
    ago: ``version``. I decided to create a complex ID to identify the data on
    remote machine.:

    <path>/<version>/<packetName> This way same data can have
    different versions. See ``lib.packetid`` module for more info.
    """
    return cmp(modified_version(version1), modified_version(version2))


def backup_id_compare(backupID1, backupID2):
    """
    Compare two 'complex' backupID's: at first compare paths, than version.
    """
    if isinstance(backupID1, tuple):
        backupID1 = backupID1[0]
        backupID2 = backupID2[0]
    customerGlobalID1, remotePath1, version1 = packetid.SplitBackupID(backupID1)
    customerGlobalID2, remotePath2, version2 = packetid.SplitBackupID(backupID2)
    if remotePath1 is None or remotePath2 is None:
        return 0
    if remotePath1 != remotePath2:
        return cmp(remotePath1, remotePath2)
    if customerGlobalID1 != customerGlobalID2:
        return cmp(customerGlobalID1, customerGlobalID2)
    return version_compare(version1, version2)


def sorted_backup_ids(backupIds, reverse=False):
    """
    Sort a list of backupID's.
    """
    sorted_ids = sorted(backupIds, key=functools.cmp_to_key(backup_id_compare), reverse=reverse)
    return sorted_ids


def sorted_versions(versions, reverse=False):
    """
    Sort a list of versions.
    """
    sorted_versions_list = sorted(versions, key=functools.cmp_to_key(version_compare), reverse=reverse)
    return sorted_versions_list


#------------------------------------------------------------------------------


def DigitsOnly(inpt, includes=''):
    """
    Very basic method to convert string to number.

    This returns same string but with digits only.
    """
    return ''.join([c for c in inpt if c in '0123456789' + includes])


def IsDigitsOnly(inpt):
    """
    Return True if ``input`` string contains only digits.
    """
    for c in inpt:
        if c not in '0123456789':
            return False
    return True


def ToInt(inpt, default=0):
    """
    Convert a string to number using built-in int() method.
    """
    try:
        return int(inpt)
    except:
        return default


def ToFloat(inpt, default=0.0):
    """
    Convert a string to number using built-in float() method.
    """
    try:
        return float(inpt)
    except:
        return default


#------------------------------------------------------------------------------


def ValidKeyAlias(key_alias):
    if len(key_alias) > 50:
        lg.warn('key_alias is too long')
        return False
    if len(key_alias) < settings.MinimumUsernameLength():
        lg.warn('key_alias is too short')
        return False
    pos = 0
    for c in key_alias:
        if c not in settings.LegalUsernameChars():
            lg.warn('key_alias has illegal character at position: %d' % pos)
            return False
        pos += 1
    if key_alias[0] not in set('abcdefghijklmnopqrstuvwxyz'):
        lg.warn('key_alias not begins with letter')
        return False
    return True


def ValidUserName(username):
    """
    A method to validate account name entered by user.
    """
    if len(username) < settings.MinimumUsernameLength():
        lg.warn('username is too short')
        return False
    if len(username) > settings.MaximumUsernameLength():
        lg.warn('username is too long')
        return False
    pos = 0
    for c in username:
        if c not in settings.LegalUsernameChars():
            lg.warn('username has illegal character at position: %d' % pos)
            return False
        pos += 1
    if username[0] not in set('abcdefghijklmnopqrstuvwxyz'):
        lg.warn('username not begins with letter')
        return False
    return True


def ValidNickName(username):
    """
    A method to validate account name entered by user.
    """
    if len(username) < settings.MinimumUsernameLength():
        return False
    if len(username) > settings.MaximumUsernameLength():
        return False
    for c in username:
        if c not in settings.LegalNickNameChars():
            return False
    return True


def ValidEmail(email, full_check=True):
    """
    A method to validate typed email address.
    """
    regexp = '^[\w\-\.\@]*$'
    if re.match(regexp, email) is None:
        return False
    if email.startswith('.'):
        return False
    if email.endswith('.'):
        return False
    if email.startswith('-'):
        return False
    if email.endswith('-'):
        return False
    if email.startswith('@'):
        return False
    if email.endswith('@'):
        return False
    if len(email) < 3:
        return False
    if full_check:
        regexp2 = '^[\w\-\.]*\@[\w\-\.]*$'
        if re.match(regexp2, email) is None:
            return False
    return True


def ValidPhone(value):
    """
    A method to validate typed phone number.
    """
    regexp = '^[ \d\-\+]*$'
    if re.match(regexp, value) is None:
        return False
    if len(value) < 5:
        return False
    return True


def ValidName(value):
    """
    A method to validate user name.
    """
    regexp = '^[\w\-]*$'
    if re.match(regexp, value) is None:
        return False
    if len(value) > 100:
        return False
    return True


def MakeValidHTMLComment(text):
    """
    Keeps only ascii symbols of the string.
    """
    ret = ''
    for c in text:
        if c in set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-+*/=_()[]{}:;,.?!@#$%|~ '):
            ret += c
    return ret


def ValidateBitCoinAddress(strAddr):
    """
    Does simple validation of a bitcoin address.

        :param strAddr: an ASCII or unicode string, of a bitcoin public address.
        :return boolean: indicating that the address has a correct format.
    http://www.rugatu.com/questions/3255/anybody-has-python-code-to-verifyvalidate-bitcoin-address
    """
    # The first character indicates the "version" of the address.
    CHARS_OK_FIRST = '123'
    # alphanumeric characters without : l I O 0
    CHARS_OK = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
    # We do not check the high length limit of the adress.
    if len(strAddr) < 27:
        return False
    if len(strAddr) > 35:
        return False
    if strAddr[0] not in CHARS_OK_FIRST:
        return False
    # We use the function "all" by passing it an enumerator as parameter.
    # It does a little optimisation :
    # if one of the character is not valid, the next ones are not tested.
    return all((char in CHARS_OK for char in strAddr[1:]))


#------------------------------------------------------------------------------


def RoundupFile(filename, stepsize):
    return local_fs.RoundupFile(filename=filename, stepsize=stepsize)


def RoundupString(data, stepsize):
    size = len(data)
    mod = size % stepsize
    increase = 0
    addon = ''
    if mod > 0:
        increase = stepsize - mod
        addon = ' '*increase
    return data + addon


def AddNL(s):
    """
    Just return a same string but with '\n' symbol at the end.
    """
    return s + '\n'


def Data():
    """
    An alias for Data packets.
    """
    return 'Data'


def Parity():
    """
    An alias for Parity packets.
    """
    return 'Parity'


#------------------------------------------------------------------------------


def pack_url_param(s):
    """
    A wrapper for built-in ``urllib.quote`` method.
    """
    try:
        return six.moves.urllib.parse.quote(s)
    except:
        try:
            return str(six.moves.urllib.parse.quote(str(s)))
        except:
            lg.exc()
    return s


def unpack_url_param(s, default=None):
    """
    A wrapper for built-in ``urllib.unquote`` method.
    """
    if s is None or not s:
        if default is not None:
            return default
        return s
    try:
        return six.moves.urllib.parse.unquote(str(s))
    except:
        lg.exc()
        return default


#------------------------------------------------------------------------------


def rndstr(length):
    """
    This generates a random string of given ``length`` - with only digits and letters.
    """
    return ''.join([random.choice(string.letters + string.digits) for i in range(0, length)])  # @UndefinedVariable


def stringToLong(s):
    """
    Not used.
    """
    return int('\0' + s, 256)


def longToString(n):
    """
    Not used.
    """
    s = n.tostring()
    if s[0] == '\0' and s != '\0':
        s = s[1:]
    return s


def receiptIDstr(receipt_id):
    """
    This method is used to make good string for receipt ID.
    """
    try:
        return '%08d' % int(receipt_id)
    except:
        return str(receipt_id)


def username2idurl(username, host='id.bitdust.io'):
    """
    Creates an IDURL from given username, default identity server is used.
    """
    return 'http://' + host + '/' + username + '.xml'


def calculate_best_dimension(sz, maxsize=8):
    """
    This method is used to visually organize users on screen. Say 4 items is
    pretty good looking in one line. But 13 items seems fine in three lines.

    :param sz: number of items to be organized
    :param maxsize: the maximum width of the matrix.
    """
    cached = {2: (2, 1), 4: (4, 1), 7: (4, 2), 13: (5, 3), 18: (6, 3), 26: (7, 4), 64: (8, 8)}.get(sz, None)
    if cached:
        return cached
    try:
        w = math.sqrt(sz)
        h = sz/w
    except:
        lg.exc()
    w = w*1.4
    h = h/1.4
    if int(w)*int(h) < sz and int(h) > 0:
        w += 1.0
    if w > maxsize:
        w = float(maxsize)
        h = sz/w
    w = int(w)
    h = int(h)
    w = 1 if w == 0 else w
    h = 1 if h == 0 else h
    if w*h < sz:
        h += 1
    if w*h - sz > h:
        w -= 1
    return w, h


def calculate_padding(w, h):
    """
    Calculates space between icons to show in the GUI.

    Need to put less spaces when show a lot of items.
    """
    imgW = 64
    imgH = 64
    if w >= 4:
        imgW = 4*imgW/w
        imgH = 4*imgH/w
    padding = 64/w - 8
    return imgW, imgH, padding


def getDeltaTime(tm):
    """
    Return a string shows how much time passed since ``tm`` moment.
    """
    try:
        #        tm = time.mktime(time.strptime(self.backupID, "F%Y%m%d%I%M%S%p"))
        dt = round(time.time() - tm)
        if dt > 2*60*60:
            return round(dt/(60.0*60.0)), 'hours'
        if dt > 60:
            return round(dt/60.0), 'minutes'
        return dt, 'seconds'
    except:
        return None, None


def getRealHost(host, port=None):
    """
    Some tricks to get a 'host' from contact method (see ``lib.identity``).
    """
    if isinstance(host, six.string_types):
        if port is not None:
            host += ':' + str(port)
    elif isinstance(host, tuple) and len(host) == 2:
        host = host[0] + ':' + str(host[1])
    elif host is None:
        host = 'None'
    else:
        if getattr(host, 'host', None) is not None:
            if getattr(host, 'port', None) is not None:
                host = str(getattr(host, 'host')) + ':' + str(getattr(host, 'port'))
            else:
                host = str(getattr(host, 'host'))
        elif getattr(host, 'underlying', None) is not None:
            host = str(getattr(host, 'underlying'))
        else:
            host = str(host)
            if port is not None:
                host += ':' + str(port)
    return host


def split_geom_string(geomstr):
    """
    Split strings created with format "%dx%d+%d+%d" into 4 integers.
    """
    try:
        r = re.split('\D+', geomstr, 4)
        return int(r[0]), int(r[1]), int(r[2]), int(r[3])
    except:
        return None, None, None, None


def percent2string(percent, precis=3):
    """
    A tool to make a string (with % at the end) from given float, ``precis`` is
    precision to round the number.
    """
    s = float2str(round(percent, precis), mask=('%%3.%df' % (precis + 2)))
    return s + '%'


def value2percent(value, total, precis=3):
    if not total:
        return '0%'
    return percent2string(100.0*(float(value)/float(total)), precis)


def float2str(float_value, mask='%6.8f', no_trailing_zeros=True):
    """
    Some smart method to do simple operation - convert float value into string.
    """
    try:
        f = float(float_value)
    except:
        return float_value
    s = mask % f
    if no_trailing_zeros:
        s = s.rstrip('0').rstrip('.')
    return s


def seconds_to_time_left_string(seconds):
    """
    Using this method you can print briefly some period of time.

    This is my post on StackOverflow to share that:
    http://stackoverflow.com/questions/538666/python-format-timedelta-
    to-string/19074707#19074707
    """
    s = int(seconds)
    years = s // 31104000
    if years > 1:
        return '%d years' % years
    s = s - (years*31104000)
    months = s // 2592000
    if years == 1:
        r = 'one year'
        if months > 0:
            r += ' and %d months' % months
        return r
    if months > 1:
        return '%d months' % months
    s = s - (months*2592000)
    days = s // 86400
    if months == 1:
        r = 'one month'
        if days > 0:
            r += ' and %d days' % days
        return r
    if days > 1:
        return '%d days' % days
    s = s - (days*86400)
    hours = s // 3600
    if days == 1:
        r = 'one day'
        if hours > 0:
            r += ' and %d hours' % hours
        return r
    s = s - (hours*3600)
    minutes = s // 60
    seconds = s - (minutes*60)
    if hours >= 6:
        return '%d hours' % hours
    if hours >= 1:
        r = '%d hours' % hours
        if hours == 1:
            r = 'one hour'
        if minutes > 0:
            r += ' and %d minutes' % minutes
        return r
    if minutes == 1:
        r = 'one minute'
        if seconds > 0:
            r += ' and %d seconds' % seconds
        return r
    if minutes == 0:
        return '%d seconds' % seconds
    if seconds == 0:
        return '%d minutes' % minutes
    return '%d minutes and %d seconds' % (minutes, seconds)


def unicode_to_str_safe(unicode_string, encodings=None):
    """
    I tried to make an 'ultimate' method to convert unicode to string here.
    """
    try:
        return str(unicode_string)  # .decode('utf-8')
    except:
        try:
            return six.text_type(unicode_string).encode(locale.getpreferredencoding(), errors='ignore')
        except:
            pass
    if encodings is None:
        encodings = [
            locale.getpreferredencoding(),
        ]  # 'utf-8'
    output = ''
    for i in range(len(unicode_string)):
        unicode_char = unicode_string[i]
        char = '?'
        try:
            char = unicode_char.encode(encodings[0])
            # print char, encodings[0]
        except:
            for encoding in encodings:
                try:
                    char = unicode_char.encode(encoding, errors='ignore')
                    # print char, encoding
                    break
                except:
                    pass
        output += char
    return output


def wrap_long_string(longstring, width=40, wraptext='\n'):
    w = len(longstring)
    if w < width:
        return longstring
    return wraptext.join(textwrap.wrap(longstring, width))


def cut_long_string(longstring, length=40, suffix=''):
    l = len(longstring)
    if l < length:
        return longstring
    return longstring[:length] + suffix


def isEnglishString(s):
    try:
        s.decode('ascii')
    except UnicodeDecodeError:
        return False
    else:
        return True


#------------------------------------------------------------------------------


def getClipboardText():
    """
    A portable way to get a clipboard data - some sort of Ctrl-V.
    """
    if bpio.Windows():
        try:
            import win32clipboard  # @UnresolvedImport
            import win32con  # @UnresolvedImport
            win32clipboard.OpenClipboard()
            d = win32clipboard.GetClipboardData(win32con.CF_TEXT)
            win32clipboard.CloseClipboard()
            return d.replace('\r\n', '\n')
        except:
            lg.exc()
            return ''
    elif bpio.Linux():
        try:
            import wx
            # may crash, otherwise
            # this needs app.MainLoop() to be started
            if not wx.TheClipboard.IsOpened():  # @UndefinedVariable
                do = wx.TextDataObject()  # @UndefinedVariable
                wx.TheClipboard.Open()  # @UndefinedVariable
                success = wx.TheClipboard.GetData(do)  # @UndefinedVariable
                wx.TheClipboard.Close()  # @UndefinedVariable
                if success:
                    return do.GetText()
                else:
                    return ''
            else:
                return ''
        except:
            return ''
    else:
        return ''


def setClipboardText(txt):
    """
    A portable way to set a clipboard data - just like when you select something and press Ctrl-C.
    """
    if bpio.Windows():
        try:
            import win32clipboard  # @UnresolvedImport
            import win32con  # @UnresolvedImport
            win32clipboard.OpenClipboard()
            win32clipboard.EmptyClipboard()
            win32clipboard.SetClipboardData(win32con.CF_TEXT, txt)
            win32clipboard.CloseClipboard()
        except:
            lg.exc()

    elif bpio.Linux():
        try:
            import wx
            clipdata = wx.TextDataObject()  # @UndefinedVariable
            clipdata.SetText(txt)
            if wx.TheClipboard:  # @UndefinedVariable
                wx.TheClipboard.Open()  # @UndefinedVariable
                wx.TheClipboard.SetData(clipdata)  # @UndefinedVariable
                wx.TheClipboard.Close()  # @UndefinedVariable
        except:
            lg.exc()

    elif bpio.Mac():
        try:
            fd, fname = tempfile.mkstemp()
            os.write(fd, txt)
            os.close(fd)
            os.system('cat %s | pbcopy' % fname)
            os.remove(fname)
        except:
            lg.exc()


#------------------------------------------------------------------------------


def encode64(s):
    """
    A wrapper for built-in ``base64.b64encode``.
    """
    return base64.b64encode(s)


def decode64(s):
    """
    A wrapper for built-in ``base64.b64decode``.
    """
    return base64.b64decode(s)


def get_hash(src):
    """
    Get a good looking MD5 hash of ``src`` string.
    """
    return hashlib.md5(src).hexdigest()


def file_hash(path):
    """
    Read file and get get its hash.
    """
    src = bpio.ReadBinaryFile(path)
    if not src:
        return None
    return get_hash(src)


#-------------------------------------------------------------------------------


def time2daystring(tm=None):
    """
    Use built-in method ``time.strftime`` to conver ``tm`` to string in
    '%Y%m%d' format.
    """
    tm_ = tm
    if tm_ is None:
        tm_ = time.time()
    return time.strftime('%Y%m%d', time.localtime(tm_))


def daystring2time(daystring):
    """
    Reverse method for ``time2daystring``.
    """
    try:
        t = time.strptime(daystring, '%Y%m%d')
    except:
        return None
    return time.mktime(t)


def time2str(format):
    """
    A wrapper for ``time.strftime``.
    """
    return time.strftime(format)


def gmtime2str(format, seconds=None):
    """
    Almost the same to ``time2str``, but uses ``time.gmtime`` to get the
    current moment.
    """
    if not seconds:
        return time.strftime(format, time.gmtime())
    return time.strftime(format, time.gmtime(seconds))


def str2gmtime(time_string, format):
    """
    A reverse method for ``gmtime2str``.
    """
    return time.mktime(time.strptime(time_string, format))


#------------------------------------------------------------------------------


def DoRestart(param='', detach=False, std_out='/dev/null', std_err='/dev/null'):
    """
    A smart and portable way to restart a whole program.
    """
    if bpio.Windows():
        if bpio.isFrozen():
            # lg.out(2, "misc.DoRestart under Windows (Frozen), param=%s" % param)
            # lg.out(2, "misc.DoRestart sys.executable=" + sys.executable)
            # lg.out(2, "misc.DoRestart sys.argv=" + str(sys.argv))
            starter_filepath = os.path.join(bpio.getExecutableDir(), settings.WindowsStarterFileName())
            if not os.path.isfile(starter_filepath):
                # lg.out(2, "misc.DoRestart ERROR %s not found" % starter_filepath)
                main_filepath = os.path.join(bpio.getExecutableDir(), settings.WindowsMainScriptFileName())
                cmdargs = [
                    os.path.basename(main_filepath),
                ]
                if param != '':
                    cmdargs.append(param)
                # lg.out(2, "misc.DoRestart cmdargs="+str(cmdargs))
                return os.spawnve(os.P_DETACH, main_filepath, cmdargs, os.environ)  # @UndefinedVariable
            cmdargs = [
                os.path.basename(starter_filepath),
            ]
            if param != '':
                cmdargs.append(param)
            # lg.out(2, "misc.DoRestart cmdargs="+str(cmdargs))
            return os.spawnve(os.P_DETACH, starter_filepath, cmdargs, os.environ)  # @UndefinedVariable

        pypath = sys.executable
        cmdargs = [
            sys.executable,
        ]
        cmdargs.append(sys.argv[0])
        cmdargs += sys.argv[1:]
        if param != '' and not sys.argv.count(param):
            cmdargs.append(param)
        if cmdargs.count('restart'):
            cmdargs.remove('restart')
        if cmdargs.count('detach'):
            cmdargs.remove('detach')
        if cmdargs.count('daemon'):
            cmdargs.remove('daemon')
        if detach:
            from bitdust.system import child_process
            cmdargs = [strng.to_text(a) for a in cmdargs]
            return child_process.detach(cmdargs)
        return os.execvpe(pypath, cmdargs, os.environ)

    pypyth = sys.executable
    cmdargs = [
        sys.executable,
    ]
    if sys.argv[0] == '/usr/share/bitdust/bitdust.py':
        cmdargs.append('/usr/bin/bitdust')
    else:
        cmdargs.append(sys.argv[0])
    if param:
        cmdargs.append(param)
    if cmdargs.count('restart'):
        cmdargs.remove('restart')
    if cmdargs.count('detach'):
        cmdargs.remove('detach')
    if cmdargs.count('daemon'):
        cmdargs.remove('daemon')
    pid = os.fork()
    if pid != 0:
        return None
    if detach:
        cmdargs[1] = os.path.abspath(cmdargs[1])
        cmdargs.append('1>%s' % std_out)
        cmdargs.append('2>%s' % std_err)
        cmd = '/usr/bin/nohup ' + (' '.join(cmdargs)) + ' &'
        BITDUST_COVERAGE_PROCESS_START = os.environ.get('COVERAGE_PROCESS_START')
        if BITDUST_COVERAGE_PROCESS_START:
            cmd = 'COVERAGE_PROCESS_START="%s" %s' % (
                BITDUST_COVERAGE_PROCESS_START,
                cmd,
            )
        BITDUST_LOG_USE_COLORS = os.environ.get('BITDUST_LOG_USE_COLORS')
        if BITDUST_LOG_USE_COLORS:
            cmd = 'BITDUST_LOG_USE_COLORS="%s" %s' % (
                BITDUST_LOG_USE_COLORS,
                cmd,
            )
        return os.system(cmd)
    return os.execvpe(pypyth, cmdargs, os.environ)


def ExplorePathInOS(filepath):
    """
    Very nice and portable way to show location or file on local disk.
    """
    try:
        if bpio.Windows():
            # os.startfile(filepath)
            if os.path.isfile(filepath):
                subprocess.Popen(['explorer', '/select,', '%s' % (filepath.replace('/', '\\'))])
            else:
                subprocess.Popen(['explorer', '%s' % (filepath.replace('/', '\\'))])

        elif bpio.Linux():
            subprocess.Popen(['`which xdg-open`', filepath])

        elif bpio.Mac():
            subprocess.Popen(['open', '-R', filepath])

    except:
        try:
            import webbrowser
            webbrowser.open(filepath)
        except:
            lg.exc()
    return


#------------------------------------------------------------------------------


def LoopAttenuation(current_delay, go_faster, min_delay, max_delay):
    """
    Pretty common method. Twisted reactor is very nice, you can call
    ``reactor.callLater(3, method_a, 'param1')`` and method_a('param1') will be
    called exactly when 3 seconds passed. But we do not want fixed periods
    sometimes.

    It is common to be in hurry when you have a lot of work, but when there is nothing to do you can take rest.
    For example, while reading a queue of events it make sense to do it as fast as possible, but only when you have some items in the queue.
    When the queue is empty - you can try to reduce the load on the main process and just check the queue less often.
    This method is used to calculate the delay to the next call of some 'idle' method.
        :param current_delay: current period of time in seconds between calls
        :param go_faster:  if this is True - method should return ``min`` period - call next time as soon as possible
                        if this is False - method will multiply ``current_delay`` by ``_AttenuationFactor`` and so decrease the speed
        :param min_delay: the minimum delay between calls
        :param max_delay: the maximum delay between calls
    """
    global _AttenuationFactor
    if go_faster:
        return min_delay
    if current_delay < max_delay:
        current_delay *= _AttenuationFactor
        if current_delay > max_delay:
            current_delay = max_delay
    return current_delay


#------------------------------------------------------------------------------

if __name__ == '__main__':
    lg.set_debug_level(10)
    bpio.init()
    init()



#!/usr/bin/env python
# file_up.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (file_up.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""

.. raw:: html

    <a href="https://bitdust.io/automats/customer/file_up.png" target="_blank">
    <img src="https://bitdust.io/automats/customer/file_up.png" style="max-width:100%;">
    </a>


.. module:: file_up
.. role:: red


BitDust file_up() Automat

EVENTS:
    * :red:`ack-received`
    * :red:`data-sent`
    * :red:`error`
    * :red:`fail-received`
    * :red:`file-not-exist`
    * :red:`init`
    * :red:`sending-failed`
    * :red:`start`
    * :red:`stop`
    * :red:`timeout`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

#------------------------------------------------------------------------------

import os
import sys
import time

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in file_up.py')

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import utime
from bitdust.lib import packetid
from bitdust.lib import nameurl

from bitdust.system import bpio

from bitdust.userid import global_id

from bitdust.main import settings

from bitdust.p2p import p2p_service
from bitdust.p2p import commands

from bitdust.transport import packet_out

from bitdust.stream import io_throttle

#------------------------------------------------------------------------------


class FileUp(automat.Automat):
    """
    This class implements all the functionality of ``file_up()`` state machine.
    """
    def __init__(self, parent, fileName, packetID, remoteID, ownerID, callOnAck=None, callOnFail=None, debug_level=_DebugLevel, log_events=_Debug, log_transitions=_Debug, publish_events=False, **kwargs):
        """
        Builds `file_up()` state machine.
        """
        self.parent = parent
        self.fileName = fileName
        try:
            self.fileSize = os.path.getsize(os.path.abspath(fileName))
        except:
            lg.exc()
            self.fileSize = 0
        self.packetID = global_id.CanonicalID(packetID)
        parts = global_id.NormalizeGlobalID(packetID)
        self.customerID = parts['customer']
        self.remotePath = parts['path']
        self.customerIDURL = parts['idurl']
        customerGlobalID, remotePath, versionName, fileName = packetid.SplitVersionFilename(packetID)
        self.backupID = packetid.MakeBackupID(customerGlobalID, remotePath, versionName)
        self.remoteID = remoteID
        self.ownerID = ownerID
        self.callOnAck = callOnAck
        self.callOnFail = callOnFail
        self.sendTime = None
        self.ackTime = None
        self.sendTimeout = 10*2*(max(int(self.fileSize/settings.SendingSpeedLimit()), 5) + 5)  # maximum 5 seconds to get an Ack
        self.result = ''
        self.created = utime.utcnow_to_sec1970()
        super(FileUp, self).__init__(
            name='file_up_%s_%s/%s/%s' % (nameurl.GetName(self.remoteID), remotePath, versionName, fileName), state='AT_STARTUP', debug_level=debug_level, log_events=log_events, log_transitions=log_transitions,
            publish_events=publish_events, **kwargs
        )

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'IN_QUEUE'
                self.doInit(*args, **kwargs)
                self.doQueueAppend(*args, **kwargs)
        #---IN_QUEUE---
        elif self.state == 'IN_QUEUE':
            if event == 'stop':
                self.state = 'STOPPED'
                self.doQueueRemove(*args, **kwargs)
                self.doReportStopped(*args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'start':
                self.state = 'UPLOADING'
                self.doSendData(*args, **kwargs)
            elif event == 'file-not-exist':
                self.state = 'NO_FILE'
                self.doCancelPackets(*args, **kwargs)
                self.doQueueRemove(*args, **kwargs)
                self.doReportFailed(event, *args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---UPLOADING---
        elif self.state == 'UPLOADING':
            if event == 'stop':
                self.state = 'STOPPED'
                self.doCancelPackets(*args, **kwargs)
                self.doQueueRemove(*args, **kwargs)
                self.doReportStopped(*args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'ack-received':
                self.state = 'DELIVERED'
                self.doQueueRemove(*args, **kwargs)
                self.doReportDelivered(*args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'timeout' or event == 'fail-received' or event == 'error' or event == 'sending-failed':
                self.state = 'FAILED'
                self.doCancelPackets(*args, **kwargs)
                self.doQueueRemove(*args, **kwargs)
                self.doReportFailed(event, *args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'data-sent':
                self.state = 'ACK?'
        #---DELIVERED---
        elif self.state == 'DELIVERED':
            pass
        #---STOPPED---
        elif self.state == 'STOPPED':
            pass
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        #---NO_FILE---
        elif self.state == 'NO_FILE':
            pass
        #---ACK?---
        elif self.state == 'ACK?':
            if event == 'stop':
                self.state = 'STOPPED'
                self.doCancelPackets(*args, **kwargs)
                self.doQueueRemove(*args, **kwargs)
                self.doReportStopped(*args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'ack-received':
                self.state = 'DELIVERED'
                self.doQueueRemove(*args, **kwargs)
                self.doReportDelivered(*args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'error' or event == 'timeout' or event == 'fail-received' or event == 'sending-failed':
                self.state = 'FAILED'
                self.doCancelPackets(*args, **kwargs)
                self.doQueueRemove(*args, **kwargs)
                self.doReportFailed(event, *args, **kwargs)
                self.doQueueNext(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        io_throttle.PacketReport('send', self.remoteID, self.packetID, 'init')

    def doQueueAppend(self, *args, **kwargs):
        """
        Action method.
        """
        if self.packetID in self.parent.fileSendQueue:
            raise Exception('file %r already in uploading queue for %r' % (self.packetID, self.remoteID))
        if self.packetID in self.parent.fileSendDict:
            raise Exception('file %r already in uploading dict for %r' % (self.packetID, self.remoteID))
        self.parent.fileSendQueue.append(self.packetID)
        self.parent.fileSendDict[self.packetID] = self

    def doQueueRemove(self, *args, **kwargs):
        """
        Action method.
        """
        if self.packetID not in self.parent.fileSendDict:
            raise Exception('file %r not found in uploading dict for %r' % (self.packetID, self.remoteID))
        if self.packetID not in self.parent.fileSendQueue:
            raise Exception('file %r not found in uploading queue for %r' % (self.packetID, self.remoteID))
        self.parent.fileSendQueue.remove(self.packetID)
        del self.parent.fileSendDict[self.packetID]

    def doSendData(self, *args, **kwargs):
        """
        Action method.
        """
        payload = bpio.ReadBinaryFile(self.fileName)
        if not payload:
            self.event('error', Exception('file %r reading error' % self.fileName))
            return
        p2p_service.SendData(
            raw_data=payload,
            ownerID=self.ownerID,
            creatorID=self.parent.creatorID,
            remoteID=self.remoteID,
            packetID=self.packetID,
            callbacks={
                commands.Ack(): self.parent.OnFileSendAckReceived,
                commands.Fail(): self.parent.OnFileSendAckReceived,
            },
        )
        self.sendTime = time.time()

    def doCancelPackets(self, *args, **kwargs):
        """
        Action method.
        """
        packetsToCancel = packet_out.search_by_packet_id(self.packetID)
        for pkt_out in packetsToCancel:
            if pkt_out.outpacket.Command == commands.Data():
                if _Debug:
                    lg.dbg(_DebugLevel, 'sending "cancel" to %s addressed to %s because downloading cancelled' % (pkt_out, pkt_out.remote_idurl))
                pkt_out.automat('cancel')

    def doQueueNext(self, *args, **kwargs):
        """
        Action method.
        """
        self.parent.DoSend()

    def doReportDelivered(self, *args, **kwargs):
        """
        Action method.
        """
        self.ackTime = time.time()
        self.parent.uploadingTimeoutCount = 0
        if self.callOnAck:
            newpacket = args[0]
            reactor.callLater(0, self.callOnAck, newpacket, newpacket.OwnerID, self.packetID)  # @UndefinedVariable

    def doReportStopped(self, *args, **kwargs):
        """
        Action method.
        """
        if self.callOnFail:
            reactor.callLater(0, self.callOnFail, self.remoteID, self.packetID, 'failed')  # @UndefinedVariable

    def doReportFailed(self, event, *args, **kwargs):
        """
        Action method.
        """
        if event == 'fail-received':
            if self.callOnFail:
                reactor.callLater(0, self.callOnFail, self.remoteID, self.packetID, 'failed')  # @UndefinedVariable
        elif event == 'timeout':
            self.parent.uploadingTimeoutCount += 1
            if self.callOnFail:
                reactor.callLater(0, self.callOnFail, self.remoteID, self.packetID, 'timeout')  # @UndefinedVariable
        elif event == 'sending-failed':
            if self.callOnFail:
                reactor.callLater(0, self.callOnFail, self.remoteID, self.packetID, 'failed')  # @UndefinedVariable
        else:
            if self.callOnFail:
                reactor.callLater(0, self.callOnFail, self.remoteID, self.packetID, 'failed')  # @UndefinedVariable

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        io_throttle.PacketReport('send', self.remoteID, self.packetID, self.result)
        self.parent = None
        self.fileName = None
        self.fileSize = None
        self.packetID = None
        self.customerID = None
        self.remotePath = None
        self.customerIDURL = None
        self.backupID = None
        self.remoteID = None
        self.ownerID = None
        self.callOnAck = None
        self.callOnFail = None
        self.sendTime = None
        self.ackTime = None
        self.sendTimeout = None
        self.result = None
        self.created = None
        self.destroy()



#!/usr/bin/env python
# utime.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (utime.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
..

module:: utime
"""

from __future__ import absolute_import

#------------------------------------------------------------------------------

import sys
import time
import datetime
import calendar

#------------------------------------------------------------------------------

if sys.version_info < (3, 7):

    def old_fromisoformat(text):
        """ for python versions < 3.7 get datetime from isoformat """
        d, t = text.split('T')
        year, month, day = d.split('-')
        hours, minutes, seconds = t.split(':')
        seconds = float(seconds[0:-1])
        sec = int(seconds)
        microseconds = int((seconds - sec)*1e6)
        return datetime.datetime(int(year), int(month), int(day), int(hours), int(minutes), sec, microseconds)

    _fromisoformat = old_fromisoformat

else:
    _fromisoformat = datetime.datetime.fromisoformat

#------------------------------------------------------------------------------


def utc_timestamp(d):
    timestamp1 = calendar.timegm(d.timetuple())
    return datetime.datetime.utcfromtimestamp(timestamp1)


def local_timestamp(d):
    timestamp2 = time.mktime(d.timetuple())
    return datetime.datetime.fromtimestamp(timestamp2)


def datetime_to_sec1970(dt):
    """
    Converts datetime object to seconds since 1970 year.
    """
    return int(calendar.timegm(dt.timetuple()))


def sec1970_to_datetime_utc(seconds=-1):
    """
    Converts seconds since 1970 year to datetime object in UTC form.
    """
    if seconds is None:
        return None
    if seconds == -1:
        seconds = utcnow_to_sec1970()
    return datetime.datetime.utcfromtimestamp(seconds)


def utcnow_to_sec1970():
    """
    Returns how much seconds passed since 1970 till current moment counting from
    UTC timezone.
    TODO: extra methods for time synchronization across the network nodes to be added later
    """
    return datetime_to_sec1970(datetime.datetime.utcnow())


def get_sec1970():
    """
    Return how much seconds passed since 1970 using time.time() method.
    TODO: extra methods for time synchronization across the network nodes to be added later
    """
    return int(time.time())


def get_milliseconds1970():
    """
    Returns how much `milliseconds` passed since 1970 till current moment.
    TODO: extra methods for time synchronization across the network nodes to be added later
    """
    return int(time.time()*1000.0)


def make_timestamp():
    """
    Returns text string based on current time.
    """
    time_st = time.localtime()
    ampm = time.strftime('%p', time_st)
    if not ampm:
        ampm = 'AM' if time.time() % 86400 < 43200 else 'PM'
    return time.strftime('%Y%m%d%I%M%S', time_st) + ampm


def pack_time(seconds=-1, timespec='seconds'):
    """
    Converts seconds since 1970 year to ISO formatted string.
    """
    if seconds is None:
        return None
    return sec1970_to_datetime_utc(seconds).isoformat(timespec=timespec)


def unpack_time(text):
    """
    Converts ISO formatted string to seconds since 1970 year.
    """
    if text is None:
        return None
    return datetime_to_sec1970(_fromisoformat(text))



#!/usr/bin/python
# measure_it.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (measure_it.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: measure_it
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

from six.moves import range

import time

#------------------------------------------------------------------------------

_CallsIndex = {}

#------------------------------------------------------------------------------


def init():
    pass


def shutdown():
    global _CallsIndex
    _CallsIndex.clear()


def make_fqn(callabl, *args, **kwargs):
    fqn = ''
    nm = callabl.__class__.__name__
    if nm == 'function' or nm == 'instancemethod':
        fqn = '{}.{}'.format(callabl.__module__, callabl.__name__)
    elif nm == 'LoopingCall':
        fqn = '{}.{}'.format(callabl.f.__module__, callabl.f.__name__)
    elif nm == 'method':
        fqn = '{}.{}'.format(callabl.__func__.__module__, callabl.__func__.__name__)
    else:
        raise Exception('unexpected callable type: %r' % callabl)
    if fqn == 'transport.gateway._call':
        fqn += '.' + args[0]
    elif fqn == 'automats.automat.timerEvent':
        fqn += '.' + callabl.a[0]
    elif fqn == 'automats.automat.event':
        fqn += '.' + args[0]
    elif fqn == 'automats.automat.automat':
        fqn += '.' + args[0]
    return fqn


def top_calls(top_size=5):
    global _CallsIndex
    keys = list(_CallsIndex.keys())
    keys.sort(key=lambda cb: -_CallsIndex[cb][1])
    s = '    cumulative time:\n'
    for i in range(0, min(top_size, len(_CallsIndex))):
        cb = keys[i]
        s += '        %f sec. with %d calls : %s\n' % (_CallsIndex[cb][1], _CallsIndex[cb][0], cb)
    keys.sort(key=lambda cb: -_CallsIndex[cb][0])
    s += '    number of calls:\n'
    for i in range(0, min(top_size, len(_CallsIndex))):
        cb = keys[i]
        s += '        %f sec. with %d calls : %s\n' % (_CallsIndex[cb][1], _CallsIndex[cb][0], cb)
    return s


def run(callabl, *args, **kwargs):
    global _CallsIndex
    fqn = make_fqn(callabl, *args, **kwargs)
    tm = float(time.time())
    try:
        result = callabl(*args, **kwargs)
    except Exception as e:
        result = e
    exec_time = float(time.time()) - tm
    if fqn not in _CallsIndex:
        _CallsIndex[fqn] = [0, 0.0]
    _CallsIndex[fqn][0] += 1
    _CallsIndex[fqn][1] += exec_time
    return result, exec_time
